"""
@author: Luis Fernando Lara Tobar
@author: Peter Corke
@author: Samuel Drew
"""

from roboticstoolbox import DHRobot, RevoluteDH


class Planar3(DHRobot):
    """
    Class that models a planar 3-link robot

    ``Planar2()`` is a class which models a 3-link planar robot and
    describes its kinematic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Planar3()
        >>> print(robot)

    Defined joint configurations are:

        - qz, zero angles, all folded up

    .. note::

      - Robot has only 3 DoF.

    .. codeauthor:: Peter Corke
    """

    def __init__(self):

        L = [
                RevoluteDH(a=1),
                RevoluteDH(a=1),
                RevoluteDH(a=1)
            ]

        super().__init__(L, name='Planar 3 link', keywords=('planar',))
        self.addconfiguration("qz", [0, 0, 0])


if __name__ == '__main__':   # pragma nocover

    robot = Planar3()
    print(robot)



"""
@author: Luis Fernando Lara Tobar
@author: Peter Corke
@author: Samuel Drew
"""

from roboticstoolbox import DHRobot, RevoluteDH
from math import pi


class Planar2(DHRobot):
    """
    Class that models a planar 2-link robot

    ``Planar2()`` is a class which models a 2-link planar robot and
    describes its kinematic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Planar2()
        >>> print(robot)

    Defined joint configurations are:

        - qz, zero angles, all folded up
        - q1, links are horizontal and vertical respectively
        - q2, links are vertical and horizontal respectively

    .. note::

      - Robot has only 2 DoF.

    .. codeauthor:: Peter Corke
    """

    def __init__(self):

        L = [
                RevoluteDH(a=1),
                RevoluteDH(a=1)
            ]

        super().__init__(L, name='Planar 2 link', keywords=('planar',))
        self.addconfiguration("qz", [0, 0])
        self.addconfiguration("q1", [0, pi/2])
        self.addconfiguration("q2", [pi/2, -pi/2])


if __name__ == '__main__':   # pragma nocover

    robot = Planar2()
    print(robot)



#!/usr/bin/env python
"""
@author: Peter Corke
"""

import numpy as np
from roboticstoolbox import DHRobot, RevoluteDH
from spatialmath import SE3


class Sawyer(DHRobot):
    """
    Class that models a Sawyer manipulator

    :param symbolic: use symbolic constants
    :type symbolic: bool

    ``Sawyer()`` is an object which models a Rethink Sawyer robot and
    describes its kinematic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Sawyer()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration

    :references:

        -`Layeghi, Daniel. “Dynamic and Kinematic Modelling of the Sawyer Arm ” Google Sites, 20 Nov. 2017 <https://sites.google.com/site/daniellayeghi/daily-work-and-writing/major-project-2>`_

    .. note:: SI units of metres are used.

    """

    def __init__(self, symbolic=False):

        if symbolic:
            import spatialmath.base.symbolic as sym
            zero = sym.zero()
            pi = sym.pi()
        else:
            from math import pi
            zero = 0.0

        deg = pi / 180
        mm = 1e-3
        
        # kinematic parameters
        a = np.r_[81, 0, 0, 0, 0, 0, 0] * mm
        d = np.r_[317, 192.5, 400, 168.5, 400, 136.3, 133.75] * mm
        alpha = [-pi/2, -pi/2, -pi/2, -pi/2, -pi/2, -pi/2, 0]

        links = []

        for j in range(6):
            link = RevoluteDH(
                d=d[j],
                a=a[j],
                alpha=alpha[j]
            )
            links.append(link)
    
        super().__init__(
            links,
            name="Sawyer",
            manufacturer="Rethink Robotics",
            keywords=('redundant', 'symbolic',),
            symbolic=symbolic
        )
    
        # zero angles
        self.addconfiguration("qz", np.array([0, 0, 0, 0, 0, 0]))


if __name__ == '__main__':    # pragma nocover

    sawyer = Sawyer(symbolic=False)
    print(sawyer)



#!/usr/bin/env python
"""
@author Jesse Haviland
"""

from abc import ABC, abstractmethod


class Connector(ABC):

    def __init__(self):
        super().__init__()

    #
    #  Basic methods to do with the state of the external program
    #

    @abstractmethod
    def launch(self):
        '''
        launch(args) launch the external program with an empty or
        specific scene as defined by args

        '''

        pass

    @abstractmethod
    def step(self):
        '''
        step(dt) triggers the external program to make a time step
        of defined time updating the state of the environment as defined by
        the robot's actions.

        The will go through each robot in the list and make them act based on
        their control type (position, velocity, acceleration, or torque). Upon
        acting, the other three of the four control types will be updated in
        the internal state of the robot object. The control type is defined
        by the robot object, and not all robot objects support all control
        types.

        '''

        pass

    @abstractmethod
    def reset(self):
        '''
        reset() triggers the external program to reset to the
        original state defined by launch

        '''

        pass

    @abstractmethod
    def restart(self):
        '''
        restart() triggers the external program to close and relaunch
        to thestate defined by launch

        '''

        pass

    @abstractmethod
    def close(self):
        '''
        close() triggers the external program to gracefully close

        '''

        pass

    #
    #  Methods to interface with the robots created in other environemnts
    #

    @abstractmethod
    def add(self):
        '''
        id = add(object) adds the object to the external environment. object must
        be of an appropriate class. This adds a object object to a list of
        objects which will act upon the step() method being called.

        '''

        pass

    @abstractmethod
    def remove(self):
        '''
        remove(id) removes the object from the external environment.

        '''

        pass

    @abstractmethod
    def hold(self):    # pragma nocover
        '''
        hold() keeps the backend open i.e. stops the program from closing once
        the main script has finished. This method may need keep an even loop
        running for the backend to keep it responsive.

        '''

        pass



#!/usr/bin/env python
"""
@author Micah Huth
"""

from vpython import vector, compound, mag, box, curve
from numpy import sign, ceil, arange
from roboticstoolbox.backends.VPython.text import update_grid_numbers
from roboticstoolbox.backends.VPython.object2d import Marker2D
from spatialmath import SE2
from collections.abc import MutableMapping


class GridMMap(MutableMapping):    # pragma nocover
    def __init__(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)


class GraphicsGrid:  # pragma nocover
    """
    This class holds the current grid displayed in the canvas

    :param scene: The scene in which to add the grid
    :type scene: class:`vpython.canvas`
    """

    def __init__(self, scene, colour=None):

        # Save the scene the grid is placed in
        self.__scene = scene
        self.__is_3d = True

        self.__relative_cam = True
        self.__num_squares = 10
        self.__scale = 1

        # Save the current camera settings
        self.camera_pos = self.__scene.camera.pos
        self.camera_axes = self.__scene.camera.axis
        self.__focal_point = [
            round(self.__scene.center.x),
            round(self.__scene.center.y),
            round(self.__scene.center.z)
        ]

        if colour is None:
            colour = [0, 0, 0]
        self._colour = colour

        # Initialise a grid object
        line_thickness = min(max(self.__scale / 25, 0.01), 5)  # 0.01 -> 5
        
        self.grid_object = GridMMap({
            'xy_plane': curve(vector(0, 0, 0),
                              color=vector(self._colour[0], self._colour[1], self._colour[2]),
                              radius=line_thickness,
                              origin=vector(0, 0, 0)),
            'xz_plane': curve(vector(0, 0, 0),
                              color=vector(self._colour[0], self._colour[1], self._colour[2]),
                              radius=line_thickness,
                              origin=vector(0, 0, 0)),
            'yz_plane': curve(vector(0, 0, 0),
                              color=vector(self._colour[0], self._colour[1], self._colour[2]),
                              radius=line_thickness,
                              origin=vector(0, 0, 0)),
            'labels': [],
        })
        self.__init_grid()

        # Bind mouse releases to the update_grid function
        self.__scene.bind('mouseup keyup', self.update_grid)

    def __init_grid(self):
        """
        Initialise the grid along the x, y, z axes.
        """
        # self.grid_object[self.__planes_idx] = self.__create_grid_objects()
        self.__create_grid_objects()

        # Update the labels instead of recreating them
        update_grid_numbers(
            self.__focal_point, self.grid_object.get('labels'),
            self.__num_squares, self.__scale, self.__is_3d, self.__scene)

    def __create_grid_objects(self):
        """
        Draw a grid along each 3D plane, that is closest to the camera.
        """
        # Create grid from 0,0,0 to positive directions with step sizes of the scale
        min_x_coord = 0
        max_x_coord = min_x_coord + round(self.__num_squares * self.__scale, 2)
        min_y_coord = 0
        max_y_coord = min_y_coord + round(self.__num_squares * self.__scale, 2)
        min_z_coord = 0
        max_z_coord = min_z_coord + round(self.__num_squares * self.__scale, 2)

        # Get all coords
        x_coords = arange(
            min_x_coord, max_x_coord + self.__scale, self.__scale)
        y_coords = arange(
            min_y_coord, max_y_coord + self.__scale, self.__scale)
        z_coords = arange(
            min_z_coord, max_z_coord + self.__scale, self.__scale)

        # If the grid has given too many objects
        if len(x_coords) > self.__num_squares + 1:
            x_coords = x_coords[0:self.__num_squares + 1]
        if len(y_coords) > self.__num_squares + 1:
            y_coords = y_coords[0:self.__num_squares + 1]
        if len(z_coords) > self.__num_squares + 1:
            z_coords = z_coords[0:self.__num_squares + 1]

        # As curve objects cannot be compounded, so must be a single entity

        line_thickness = min(max(self.__scale / 25, 0.01), 5)  # 0.01 -> 5

        # Update curve objects
        self.grid_object.get('xy_plane').radius = line_thickness
        self.grid_object.get('xy_plane').append(vector(0, 0, 0))

        self.grid_object.get('xz_plane').radius = line_thickness
        self.grid_object.get('xz_plane').append(vector(0, 0, 0))

        self.grid_object.get('yz_plane').radius = line_thickness
        self.grid_object.get('yz_plane').append(vector(0, 0, 0))

        # Zig-Zag through all of the points
        # XY plane
        for idx, x_point in enumerate(x_coords):
            if idx % 2 == 0:
                y_vals = y_coords
            else:
                y_vals = y_coords[::-1]
            for y_point in y_vals:
                self.grid_object.get('xy_plane'). \
                    append(vector(x_point, y_point, 0))

        for idx, y_point in enumerate(y_coords[::-1]):
            if idx % 2 == 0:
                x_vals = x_coords[::-1]
            else:
                x_vals = x_coords
            for x_point in x_vals:
                self.grid_object.get('xy_plane'). \
                    append(vector(x_point, y_point, 0))

        # XZ Plane
        for idx, x_point in enumerate(x_coords):
            if idx % 2 == 0:
                z_vals = z_coords
            else:
                z_vals = z_coords[::-1]
            for z_point in z_vals:
                self.grid_object.get('xz_plane'). \
                    append(vector(x_point, 0, z_point))

        for idx, z_point in enumerate(z_coords[::-1]):
            if idx % 2 == 0:
                x_vals = x_coords[::-1]
            else:
                x_vals = x_coords
            for x_point in x_vals:
                self.grid_object.get('xz_plane'). \
                    append(vector(x_point, 0, z_point))

        # YZ Plane
        for idx, y_point in enumerate(y_coords):
            if idx % 2 == 0:
                z_vals = z_coords
            else:
                z_vals = z_coords[::-1]
            for z_point in z_vals:
                self.grid_object.get('yz_plane'). \
                    append(vector(0, y_point, z_point))

        for idx, z_point in enumerate(z_coords[::-1]):
            if idx % 2 == 0:
                y_vals = y_coords[::-1]
            else:
                y_vals = y_coords
            for y_point in y_vals:
                self.grid_object.get('yz_plane'). \
                    append(vector(0, y_point, z_point))

    def __move_grid_objects(self):
        """
        Reusing the current assets, move the planes to the new origins.
        """
        camera_axes = self.camera_axes
        # Locate centre of axes
        if self.__relative_cam:
            x_origin, y_origin, z_origin = round(self.__scene.center.x, 2), \
                                           round(self.__scene.center.y, 2), \
                                           round(self.__scene.center.z, 2)
            self.__focal_point = [x_origin, y_origin, z_origin]
            # Convert focal point for 2D rendering.
            # Puts focus point in centre of the view
            if not self.__is_3d:
                self.__focal_point = [
                    val - int(self.__num_squares / 2)
                    for val in self.__focal_point]
                x_origin = self.__focal_point[0]
                y_origin = self.__focal_point[1]
                z_origin = 0
                self.__focal_point[2] = z_origin
        else:
            x_origin, y_origin, z_origin = self.__focal_point[0], \
                                           self.__focal_point[1], \
                                           self.__focal_point[2]

        #   CAMERA AXES |  DISPLAYED GRID | XZ PLANE | XY PLANE | YZ PLANE
        #      x,y,z    |      x,y,z      |   x,z    |    x,y   |    y,z
        #  -------------+-----------------+----------+----------+----------
        #      -,-,-    |      +,+,+      |   +,+    |    +,+   |    +,+
        #      -,-,+    |      +,+,-      |   +,-    |    +,+   |    +,-
        #      -,+,-    |      +,-,+      |   +,+    |    +,-   |    -,+
        #      -,+,+    |      +,-,-      |   +,-    |    +,-   |    -,-
        #      +,-,-    |      -,+,+      |   -,+    |    -,+   |    +,+
        #      +,-,+    |      -,+,-      |   -,-    |    -,+   |    +,-
        #      +,+,-    |      -,-,+      |   -,+    |    -,-   |    -,+
        #      +,+,+    |      -,-,-      |   -,-    |    -,-   |    -,-
        # min = -num_squares or 0, around the default position
        # max = +num_squares or 0, around the default position
        # e.g. at the origin, for negative axes: -10 -> 0,
        # positive axes: 0 -> 10
        min_x_coord = round(
            x_origin + (-(self.__num_squares / 2)
                        + (sign(camera_axes.x) * -1) * (
                                self.__num_squares / 2)) * self.__scale, 2)

        min_y_coord = round(
            y_origin + (-(self.__num_squares / 2)
                        + (sign(camera_axes.y) * -1) * (
                                self.__num_squares / 2)) * self.__scale, 2)

        min_z_coord = round(
            z_origin + (-(self.__num_squares / 2)
                        + (sign(camera_axes.z) * -1) * (
                                self.__num_squares / 2)) * self.__scale, 2)

        if camera_axes.x < 0:
            x_pos = x_origin
        else:
            x_pos = min_x_coord
        if camera_axes.y < 0:
            y_pos = y_origin
        else:
            y_pos = min_y_coord
        if camera_axes.z < 0:
            z_pos = z_origin
        else:
            z_pos = min_z_coord

        self.grid_object.get('xy_plane').origin = \
            vector(x_pos, y_pos, z_origin)
        self.grid_object.get('xz_plane').origin = \
            vector(x_pos, y_origin, z_pos)
        self.grid_object.get('yz_plane').origin = \
            vector(x_origin, y_pos, z_pos)

    def update_grid(self):
        """
        Update the grid axes and numbers if the camera position/rotation
        has changed.
        """
        # If invisible, skip the updating (Unnecessary)
        if not self.grid_object.get('xy_plane'). \
                visible:
            return

        # Obtain the new camera settings
        new_camera_pos = vector(self.__scene.camera.pos)
        new_camera_axes = vector(self.__scene.camera.axis)

        old_camera_pos = vector(self.camera_pos)
        old_camera_axes = vector(self.camera_axes)

        # Update old positions
        self.camera_pos = new_camera_pos
        self.camera_axes = new_camera_axes

        distance_from_center = mag(
            self.__scene.center - self.__scene.camera.pos)
        if self.__is_3d:
            new_scale = round(distance_from_center / 30.0, 1)
        else:
            new_scale = round(distance_from_center / 15.0, 1)
        if not new_scale == self.__scale:
            self.set_scale(new_scale)
            if not self.__is_3d:
                self.__is_3d = True
                self.toggle_2d_3d()

        # If camera is different to previous: update
        if (not new_camera_axes.equals(old_camera_axes)) or (
                not new_camera_pos.equals(old_camera_pos)):
            # Update grid
            self.__move_grid_objects()
            update_grid_numbers(self.__focal_point,
                                self.grid_object.get('labels'),
                                self.__num_squares,
                                self.__scale,
                                self.__is_3d,
                                self.__scene)

    def toggle_2d_3d(self):
        """
        Toggle the grid between 2D and 3D options.
        2D - XY Plane
        3D - XY, XZ, YZ Planes
        """
        # Set the new visibility
        self.__is_3d = not self.__is_3d

        # Toggle it for XZ, YZ planes
        self.grid_object.get('xz_plane').visible = \
            self.__is_3d
        self.grid_object.get('yz_plane').visible = \
            self.__is_3d

        # Toggle it for Z plane numbers
        # Index start = (num_squares + 1) (11 numbers shown for 10 squares) *
        # 2 axes + 2 letters for axes
        z_label_start = (self.__num_squares + 1) * 2 + 2
        # Index end = end of labels array
        z_label_end = len(self.grid_object.get('labels'))
        # Toggle
        for idx in range(z_label_start, z_label_end):
            self.grid_object.get('labels')[idx].visible = self.__is_3d

        self.update_grid()

    def set_visibility(self, is_visible):
        """
        Set the visibility of the grid

        :param is_visible: Boolean of whether to display the grid
        :type is_visible: `bool`
        """
        # Modify all graphics
        self.grid_object.get('xy_plane').visible = is_visible
        self.grid_object.get('xz_plane').visible = is_visible
        self.grid_object.get('yz_plane').visible = is_visible
        for number in self.grid_object.get('labels'):
            number.visible = is_visible
        # If 3D, changes are made
        # If 2D and setting off, changes are made
        # If 2D and setting on, toggle the 3D graphics (are turned on)
        if self.__is_3d is False and is_visible is True:
            self.__is_3d = True
            self.toggle_2d_3d()

    def set_relative(self, is_relative):
        """
        Set whether the grid should be locked to (0, 0, 0) or relative to
        camera focus point

        :param is_relative: Whether the camera is dynamic (True) or
            static (False)
        :type is_relative: `bool`
        """
        self.__relative_cam = is_relative
        self.update_grid()

    def set_scale(self, value):
        """
        Set the scale and redraw the grid

        :param value: The value to set the scale to
        :type value: `float`
        """
        # If invisible, skip the updating (Unnecessary)
        if not self.grid_object.get('xy_plane').visible:
            return

        value = max(min(value, 100), 0.1)  # Between 0.1 and 100
        self.__scale = value
        # Turn off grid then delete
        self.grid_object.get('xy_plane').clear()
        self.grid_object.get('xz_plane').clear()
        self.grid_object.get('yz_plane').clear()
        for text in self.grid_object.get('labels'):
            text.visible = False

        # Don't del the curve objects
        self.grid_object.labels = []
        self.__init_grid()


def create_line(pos1, pos2, scene,
                colour=None, thickness=0.01, opacity=1):  # pragma nocover
    """
    Create a line from position 1 to position 2.

    :param scene: The scene in which to draw the object
    :type scene: class:`vpython.canvas`
    :param pos1: 3D position of one end of the line.
    :type pos1: class:`vpython.vector`
    :param pos2: 3D position of the other end of the line.
    :type pos2: class:`vpython.vector`
    :param colour: RGB list to colour the line to
    :type colour: `list`
    :param thickness: Thickness of the line
    :type thickness: `float`
    :param opacity: Opacity of the line
    :type opacity: `float`
    :raises ValueError: RGB colour must be normalised between 0->1
    :raises ValueError: Thickness must be greater than 0
    :return: A box resembling a line
    :rtype: class:`vpython.box`
    """
    # Set default colour
    # Stops a warning about mutable parameter
    if colour is None:
        colour = [0, 0, 0]

    if colour[0] > 1.0 or colour[1] > 1.0 or colour[2] > 1.0 or \
            colour[0] < 0.0 or colour[1] < 0.0 or colour[2] < 0.0:
        raise ValueError("RGB values must be normalised between 0 and 1")

    if thickness < 0.0:
        raise ValueError("Thickness must be greater than 0")

    # Length of the line using the magnitude
    line_len = mag(pos2 - pos1)

    # Position of the line is the midpoint (centre) between the ends
    position = (pos1 + pos2) / 2

    # Axis direction of the line (to align the box (line) to intersect the
    # two points)
    axis_dir = pos2 - pos1

    # Return a box of thin width and height to resemble a line
    # thickness = 0.01
    return box(canvas=scene,
               pos=position,
               axis=axis_dir,
               length=line_len,
               width=thickness,
               height=thickness,
               color=vector(colour[0], colour[1], colour[2]),
               opacity=opacity)


def create_segmented_line(
        pos1, pos2, scene, segment_len,
        colour=None, thickness=0.01):  # pragma nocover
    """
    Create a dashed line from position 1 to position 2.

    :param scene: The scene in which to draw the object
    :type scene: class:`vpython.canvas`
    :param pos1: 3D position of one end of the line.
    :type pos1: class:`vpython.vector`
    :param pos2: 3D position of the other end of the line.
    :type pos2: class:`vpython.vector`
    :param colour: RGB list to colour the line to
    :type colour: `list`
    :param thickness: Thickness of the line
    :type thickness: `float`
    :param segment_len: The length of the segment, and gap between segments
    :type segment_len: `float`
    :raises ValueError: RGB colour must be normalised between 0->1
    :raises ValueError: Thickness must be greater than 0
    :return: A box resembling a line
    :rtype: class:`vpython.box`
    """
    # Set default colour
    # Stops a warning about mutable parameter
    if colour is None:
        colour = [0, 0, 0]

    if colour[0] > 1.0 or colour[1] > 1.0 or colour[2] > 1.0 or \
            colour[0] < 0.0 or colour[1] < 0.0 or colour[2] < 0.0:
        raise ValueError("RGB values must be normalised between 0 and 1")

    if thickness < 0.0:
        raise ValueError("Thickness must be greater than 0")

    # Length of the line using the magnitude
    line_len = mag(pos2 - pos1)

    # Axis direction of the line
    # (to align the box (line) to intersect the two points)
    axis_dir = pos2 - pos1
    axis_dir.mag = 1.0

    # Return a compound of boxes of thin width and height to
    # resemble a dashed line
    dash_positions = []
    boxes = []
    # Translate centre pos to centre of where dashes will originate from
    pos1 = pos1 + (axis_dir * segment_len / 2)

    # Range = number of dashes (vis and invis)
    for idx in range(0, int(ceil(line_len / (segment_len / axis_dir.mag)))):
        # Add every even point (zeroth, second...) to skip gaps between boxes
        if idx % 2 == 0:
            dash_positions.append(pos1)
        pos1 = (pos1 + axis_dir * segment_len)
        # If the axis between points changes, then the line has surpassed
        # the end point. The line is done
        check_dir = pos2 - pos1
        check_dir.mag = 1.0
        if not vectors_approx_equal(axis_dir, check_dir):
            break

    for xyz in dash_positions:
        length = segment_len
        # If the box will surpass the end point
        len_to_end = (pos2 - xyz).mag
        if len_to_end < segment_len / 2:
            # Length is equal to dist to the end * 2 (as pos is middle of box)
            length = len_to_end * 2

        boxes.append(
            box(
                canvas=scene,
                pos=xyz,
                axis=axis_dir,
                length=length,
                width=thickness,
                height=thickness,
                color=vector(colour[0], colour[1], colour[2])
            )
        )

    return compound(boxes)


def create_marker(scene, x, y, shape, colour=None):  # pragma nocover
    """
    Draw the shape at the given position

    :param scene: The scene in which to draw the object
    :type scene: class:`vpython.canvas`
    :param x: The x location to draw the object at
    :type x: `float`
    :param y: The y location to draw the object at
    :type y: `float`
    :param shape: The shape of the object to draw
    :type shape: `str`
    :param colour: The colour of the object
    :type colour: `list`
    :returns: A 2D object that has been drawn
    :rtype: class:`graphics.graphics_object2d.Object2D`
    """
    # Set default colour
    # Stops a warning about mutable parameter
    if colour is None:
        colour = [0, 0, 0]

    # Create an SE2 for the object
    obj_se2 = SE2(x=x, y=y, theta=0)

    # Create the object and return it
    return Marker2D(obj_se2, scene, shape, colour)


def vectors_approx_equal(v1, v2):  # pragma nocover
    """
    Check whether the vectors are approximately equal.
    This is used where there is VERY minor floating point differences
    that can occur in VPython.

    :param v1: Vector 1
    :type v1: class:`vpython.vector`
    :param v2: Vector 2
    :type v2: class:`vpython.vector`
    :returns: True if vectors are within tolerance
    :rtype: `bool`
    """
    return abs(v1.x - v2.x < 0.001) and abs(v1.y - v2.y < 0.001) and \
           abs(v1.z - v2.z < 0.001)



#!/usr/bin/env python
"""
@author Micah Huth
"""

from vpython import shapes, radians, extrusion, vector
from roboticstoolbox.backends.VPython.stl import import_object_from_numpy_stl


class Object2D:   # pragma nocover
    """
    This object will allow the ability to update it's pose in the scene.
    For example, it could resemble a car or drone that moves around in
    2D space.

    :param se2: The SE2 object representing position and orientation
    :type se2: class:`spatialmath.se2`
    :param scene: The scene in which to add the link
    :type scene: class:`vpython.canvas`
    :param shape: The shape of the object
    :type shape: `str`
    :param colour: The colour of the shape
    :type colour: `list`
    :raises ValueError: The shape must be in the list of possible shapes
    """
    def __init__(self, se2, scene, shape, colour):
        # Save inputs
        self.se2 = se2
        self.scene = scene
        self.shape = shape
        self.size = 0
        if colour[0] > 1.0 or colour[1] > 1.0 or colour[2] > 1.0 or \
                colour[0] < 0.0 or colour[1] < 0.0 or colour[2] < 0.0:
            raise ValueError("RGB values must be normalised between 0 and 1")
        self.colourVec = vector(colour[0], colour[1], colour[2])
        self.graphic = None

    def __create_object(self):
        """
        To be overridden by child classes
        """
        pass

    def update_pose(self, new_se2):
        """
        Update the pose of the object

        :param new_se2: The SE2 representation of the pose
        :type new_se2: class:`spatialmath.se2`
        """
        self.se2 = new_se2
        x = self.se2.t[0]
        y = self.se2.t[1]
        t = self.se2.theta
        self.graphic.pos = vector(x, y, 0)
        self.graphic.axis = vector(0, 1, 0).rotate(t)

    def update_colour(self, colour):
        """
        Update the colour of the object

        :param colour: The RGB colour
        :type colour: `list`
        """
        if colour[0] > 1.0 or colour[1] > 1.0 or colour[2] > 1.0 or \
                colour[0] < 0.0 or colour[1] < 0.0 or colour[2] < 0.0:
            raise ValueError("RGB values must be normalised between 0 and 1")
        self.graphic.color = vector(colour[0], colour[1], colour[2])
        self.colourVec = vector(colour[0], colour[1], colour[2])

    def update_visibility(self, is_visible):
        """
        Update the visibility of the object

        :param is_visible: Whether to draw or not
        :type is_visible: `bool`
        """
        self.graphic.visible = is_visible

    def update_size(self, multiply):
        """
        Update the size of the object by a multiple of the original size.

        :param multiply: A number to multiply the original size by
        :type multiply: `float`, `int`
        """
        self.graphic.size = self.size * multiply


class STL2D(Object2D):   # pragma nocover
    """
    This object is for 2D objects that contain an STL.

    :param se2: The SE2 object representing position and orientation
    :type se2: class:`spatialmath.se2`
    :param scene: The scene in which to add the link
    :type scene: class:`vpython.canvas`
    :param stl_path: The file path to the STL to apply
    :type stl_path: `str`
    :param colour: The colour of the shape
    :type colour: `list`
    """
    def __init__(self, se2, scene, stl_path, colour):
        super().__init__(se2, scene, stl_path, colour)
        self.graphic = self.__create_object()
        self.size = self.graphic.size
        self.graphic.color = self.colourVec

    def __create_object(self):
        """
        Return a compound of the loaded STL

        :return: A compound object of the triangles in the STL
        :rtype: class:`vpython.compound`
        """
        return import_object_from_numpy_stl(self.shape, self.scene)


class Marker2D(Object2D):   # pragma nocover
    """
    This class will place a marker in the given location based on the given
    marker inputs

    :param se2: The SE2 object representing position and orientation
    :type se2: class:`spatialmath.se2`
    :param scene: The scene in which to add the link
    :type scene: class:`vpython.canvas`
    :param shape: The shape of the object
    :type shape: `str`
    :param colour: The colour of the shape
    :type colour: `list`
    :raises ValueError: The shape must be in the list of possible shapes
    """
    def __init__(self, se2, scene, shape, colour):
        super().__init__(se2, scene, shape, colour)

        self.__marker_size = 0.2

        marker_styles = [
            '',  # None
            '+',  # Plus
            'o',  # Circle
            '*',  # Star
            '.',  # Dot
            'x',  # Cross
            's',  # Square
            'd',  # Diamond
            '^',  # Up triangle
            'v',  # Down triangle
            '<',  # Left triangle
            '>',  # Right triangle
            'p',  # Pentagon
            'h',  # Hexagon
        ]

        if shape not in marker_styles:
            raise ValueError(
                "The shape must be in the list of possible shapes")

        # Create the object
        self.graphic = self.__create_object()

    def __create_object(self):
        """
                Create the physical graphical object

                :returns: The graphical entity
                :rtype: class:`vpython.baseobj`
                """
        if self.shape == '':
            # 2D coords of the circle boundary
            shape_path = shapes.circle(radius=self.__marker_size / 2)
        elif self.shape == '+':
            # 2D coords of the cross boundary
            shape_path = shapes.cross(
                width=self.__marker_size, thickness=self.__marker_size / 5)
        elif self.shape == 'o':
            # 2D coords of the circle boundary
            shape_path = shapes.circle(radius=self.__marker_size / 2)
        elif self.shape == '*':
            # 2D coords of the star boundary
            shape_path = shapes.star(radius=self.__marker_size / 2, n=6)
        elif self.shape == '.':
            # 2D coords of the square boundary
            shape_path = shapes.rectangle(
                width=self.__marker_size, height=self.__marker_size)
        elif self.shape == 'x':
            # 2D coords of the cross boundary
            shape_path = shapes.cross(
                width=self.__marker_size, thickness=self.__marker_size / 5,
                rotate=radians(45))
        elif self.shape == 's':
            # 2D coords of the square boundary
            shape_path = shapes.rectangle(
                width=self.__marker_size, height=self.__marker_size,
                thickness=self.__marker_size)
        elif self.shape == 'd':
            # 2D coords of the diamond boundary
            shape_path = shapes.rectangle(
                width=self.__marker_size, height=self.__marker_size,
                thickness=self.__marker_size, rotate=radians(45))
        elif self.shape == '^':
            # 2D coords of the triangle boundary
            shape_path = shapes.triangle(
                length=self.__marker_size)
        elif self.shape == 'v':
            # 2D coords of the triangle boundary
            shape_path = shapes.triangle(
                length=self.__marker_size, rotate=radians(180))
        elif self.shape == '<':
            # 2D coords of the triangle boundary
            shape_path = shapes.triangle(
                length=self.__marker_size, rotate=radians(90))
        elif self.shape == '>':
            # 2D coords of the triangle boundary
            shape_path = shapes.triangle(
                length=self.__marker_size, rotate=radians(-90))
        elif self.shape == 'p':
            # 2D coords of the pentagon boundary
            shape_path = shapes.pentagon(
                length=self.__marker_size)
        elif self.shape == 'h':
            # 2D coords of the hexagon boundary
            shape_path = shapes.hexagon(
                length=self.__marker_size)
        # CURRENTLY UNUSED
        # elif self.__shape == 'o':
        #     # 2D coords of the octagon boundary
        #     shape_path = shapes.octagon(length=1)
        # elif self.__shape == 'r':
        #     # 2D coords of the ring boundary (with thickness = 10%)
        #     shape_path = shapes.circle(radius=0.5, thickness=0.1)
        else:
            raise ValueError("Invalid shape given")

        # Create the shape
        x = self.se2.t[0]
        y = self.se2.t[1]
        obj = extrusion(scene=self.scene,
                        path=[vector(x, y, 0.001), vector(x, y, -0.001)],
                        shape=shape_path,
                        color=self.colourVec,
                        shininess=0)
        self.size = obj.size
        if self.shape == '':
            obj.visible = False
        return obj



# Copyright (c) 2015, Open Source Robotics Foundation, Inc.
# Copyright (c) 2013, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Open Source Robotics Foundation, Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior
#       written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Authors: Stuart Glaser, William Woodall, Robert Haschke
# Maintainer: Morgan Quigley <morgan@osrfoundation.org>

from __future__ import print_function, division

import ast
import glob
import math
import os
import re
import sys
import xml.dom.minidom

from copy import deepcopy
from .color import error, warning
from .xmlutils import opt_attrs, reqd_attrs, first_child_element, \
    next_sibling_element, replace_node


_basestr = str
unicode = str
encoding = {}

# Dictionary of substitution args
substitution_args_context = {}


# Stack of currently processed files
filestack = []

# The top level directory
tld = ''


def push_file(filename):
    """
    Push a new filename to the filestack.
    Instead of directly modifying filestack, a deep-copy is created and
    modified,
    while the old filestack is returned.
    This allows to store the filestack that was active when a macro or
    property is defined
    """
    global filestack
    oldstack = filestack
    filestack = deepcopy(filestack)
    filestack.append(filename)
    return oldstack


def restore_filestack(oldstack):
    global filestack
    filestack = oldstack


def abs_filename_spec(filename_spec):
    """
    Prepend the dirname of the currently processed file
    if filename_spec is not yet absolute
    """
    if not os.path.isabs(filename_spec):
        parent_filename = filestack[-1]
        basedir = os.path.dirname(parent_filename) if parent_filename else '.'
        return os.path.join(basedir, filename_spec)
    return filename_spec  # pragma: no cover


class YamlDictWrapper(dict):  # pragma: no cover
    """Wrapper class providing dotted access to dict items"""
    def __getattr__(self, item):
        try:
            result = super(YamlDictWrapper, self).__getitem__(item)
            return YamlDictWrapper(result) \
                if isinstance(result, dict) else result
        except KeyError:
            raise XacroException("No such key: '{}'".format(item))

    __getitem__ = __getattr__


def load_yaml(filename):  # pragma: no cover
    try:
        import yaml
    except Exception:
        raise XacroException("yaml support not available; install python-yaml")

    filename = abs_filename_spec(filename)
    f = open(filename)
    oldstack = push_file(filename)
    try:
        return YamlDictWrapper(yaml.safe_load(f))
    finally:
        f.close()
        restore_filestack(oldstack)
        global all_includes
        all_includes.append(filename)


# global symbols dictionary
# taking simple security measures to forbid access to __builtins__
# only the very few symbols explicitly listed are allowed
# for discussion, see:
# http://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html
global_symbols = {
    '__builtins__': {
        k: __builtins__[k] for k in [
            'list', 'dict', 'map', 'len', 'str', 'float', 'int',
            'True', 'False', 'min', 'max', 'round']}}

# also define all math symbols and functions
global_symbols.update(math.__dict__)
# expose load_yaml and abs_filename
global_symbols.update(
    dict(load_yaml=load_yaml, abs_filename=abs_filename_spec))


class XacroException(Exception):
    """
    XacroException allows to wrap another exception (exc) and to augment
    its error message: prefixing with msg and suffixing with suffix.
    str(e) finally prints: msg str(exc) suffix
    """

    def __init__(self, msg=None, suffix=None, exc=None, macro=None):
        super(XacroException, self).__init__(msg)
        self.suffix = suffix
        self.exc = exc
        self.macros = [] if macro is None else [macro]

    def __str__(self):  # pragma: no cover
        items = [super(XacroException, self).__str__(), self.exc, self.suffix]
        return ' '.join(
            [s for s in [unicode(e) for e in items] if s not in ['', 'None']])


verbosity = 1


def check_attrs(tag, required, optional):
    """
    Helper routine to fetch required and optional attributes
    and complain about any additional attributes.
    :param tag (xml.dom.Element): DOM element node
    :param required [str]: list of required attributes
    :param optional [str]: list of optional attributes
    """
    result = reqd_attrs(tag, required)
    result.extend(opt_attrs(tag, optional))
    allowed = required + optional
    extra = [
        a for a in tag.attributes.keys()
        if a not in allowed and not a.startswith("xmlns:")]
    if extra:  # pragma: no cover
        warning(
            "%s: unknown attribute(s): %s" % (tag.nodeName, ', '.join(extra)))
        if verbosity > 0:
            print_location(filestack)
    return result


class Macro(object):
    def __init__(self):
        self.body = None  # original xml.dom.Node
        self.params = []  # parsed parameter names
        self.defaultmap = {}  # default parameter values
        self.history = []  # definition history


def eval_extension(s):  # pragma: no cover

    if s == '$(cwd)':
        return os.getcwd()

    if s.startswith('$(find'):
        return tld
        # return '..'

    if s.startswith('$(arg'):
        s = s.replace('$(arg ', '')
        s = s[:-1]
        return substitution_args_context['arg'][s]

    try:
        from roslaunch.substitution_args import resolve_args, ArgException
        from rospkg.common import ResourceNotFound
        return resolve_args(
            s, context=substitution_args_context, resolve_anon=False)
    except ImportError:

        return ''
        # raise XacroException("substitution args not supported: ", exc=e)
    except ArgException as e:
        raise XacroException("Undefined substitution argument", exc=e)
    except ResourceNotFound as e:
        raise XacroException("resource not found:", exc=e)


class Table(object):
    def __init__(self, parent=None):
        self.parent = parent
        self.table = {}
        self.unevaluated = set()  # set of unevaluated variables
        # list of currently resolved vars (to resolve recursive definitions)
        self.recursive = []
        # the following variables are for debugging / checking only
        self.depth = self.parent.depth + 1 if self.parent else 0

    @staticmethod
    def _eval_literal(value):
        if isinstance(value, _basestr):
            # remove single quotes from escaped string
            if len(value) >= 2 and value[0] == "'" and value[-1] == "'":
                return value[1:-1]
            # Try to evaluate as number literal or boolean.
            # This is needed to handle numbers in property definitions as
            # numbers, not strings.
            # python3 ignores/drops underscores in number literals
            # (due to PEP515).
            # Here, we want to handle literals with underscores as plain
            # strings.
            if '_' in value:
                return value
            # order of types is important!
            for f in [int, float, lambda x: get_boolean_value(x, None)]:
                try:
                    return f(value)
                except Exception:
                    pass
        return value

    def _resolve_(self, key):
        # lazy evaluation
        if key in self.unevaluated:
            if key in self.recursive:
                raise XacroException("recursive variable definition: %s" %
                                     " -> ".join(self.recursive + [key]))
            self.recursive.append(key)
            self.table[key] = self._eval_literal(
                eval_text(self.table[key], self))
            self.unevaluated.remove(key)
            self.recursive.remove(key)

        # return evaluated result
        value = self.table[key]
        if (verbosity > 2 and self.parent is None) or verbosity > 3:   # pragma: no cover # noqa
            print("{indent}use {key}: {value} ({loc})".format(
                indent=self.depth * ' ', key=key, value=value,
                loc=filestack[-1]), file=sys.stderr)
        return value

    def __getitem__(self, key):
        if key in self.table:
            return self._resolve_(key)
        elif self.parent:
            return self.parent[key]
        else:
            raise KeyError(key)

    def _setitem(self, key, value, unevaluated):
        if key in global_symbols:
            warning("redefining global property: %s" % key)
            print_location(filestack)

        value = self._eval_literal(value)
        self.table[key] = value
        if unevaluated and isinstance(value, _basestr):
            # literal evaluation failed: re-evaluate lazily at first access
            self.unevaluated.add(key)
        elif key in self.unevaluated:   # pragma: no cover
            # all other types cannot be evaluated
            self.unevaluated.remove(key)
        if (verbosity > 2 and self.parent is None) or verbosity > 3:   # pragma: no cover # noqa
            print("{indent}set {key}: {value} ({loc})".format(
                indent=self.depth * ' ', key=key, value=value,
                loc=filestack[-1]), file=sys.stderr)

    def __setitem__(self, key, value):
        self._setitem(key, value, unevaluated=True)

    def __contains__(self, key):
        return \
            key in self.table or \
            (self.parent and key in self.parent)

    def __str__(self):    # pragma: no cover
        s = unicode(self.table)
        if isinstance(self.parent, Table):
            s += "\n  parent: "
            s += unicode(self.parent)
        return s

    def root(self):
        p = self
        while p.parent:
            p = p.parent
        return p


class NameSpace(object):
    # dot access (namespace.property) is forwarded to getitem()
    def __getattr__(self, item):
        return self.__getitem__(item)


class PropertyNameSpace(Table, NameSpace):
    def __init__(self, parent=None):
        super(PropertyNameSpace, self).__init__(parent)


class MacroNameSpace(dict, NameSpace):
    def __init__(self, *args, **kwargs):
        super(MacroNameSpace, self).__init__(*args, **kwargs)


class QuickLexer(object):
    def __init__(self, *args, **kwargs):
        if args:
            # copy attributes + variables from other instance
            other = args[0]
            self.__dict__.update(other.__dict__)
        else:
            self.res = []
            for k, v in kwargs.items():
                self.__setattr__(k, len(self.res))
                self.res.append(re.compile(v))
        self.str = ""
        self.top = None

    def lex(self, str):
        self.str = str
        self.top = None
        self.next()

    def peek(self):
        return self.top

    def next(self):
        result = self.top
        self.top = None
        if not self.str:  # empty string
            return result
        for i in range(len(self.res)):
            m = self.res[i].match(self.str)
            if m:
                self.top = (i, m.group(0))
                self.str = self.str[m.end():]
                return result
        raise XacroException('invalid expression: ' + self.str)


all_includes = []
include_no_matches_msg = \
    """Include tag's filename spec \"{}\" matched "\" no files."""


def get_include_files(filename_spec, symbols):
    try:
        filename_spec = abs_filename_spec(eval_text(filename_spec, symbols))
    except XacroException as e:    # pragma: no cover
        if e.exc and isinstance(e.exc, NameError) and symbols is None:
            raise XacroException(
                'variable filename is supported with in-order option only')
        else:
            raise

    if re.search('[*[?]+', filename_spec):    # pragma: no cover
        # Globbing behaviour
        filenames = sorted(glob.glob(filename_spec))
        if len(filenames) == 0:
            warning(include_no_matches_msg.format(filename_spec))
    else:
        # Default behaviour
        filenames = [filename_spec]

    for filename in filenames:
        global all_includes
        all_includes.append(filename)
        yield filename


def import_xml_namespaces(parent, attributes):
    """import all namespace declarations into parent"""
    for name, value in attributes.items():
        if name.startswith('xmlns:'):
            oldAttr = parent.getAttributeNode(name)
            if oldAttr and oldAttr.value != value:
                warning("inconsistent namespace redefinitions for {name}:"   # pragma: no cover # noqa
                        "\n old: {old}\n new: {new} ({new_file})".format(
                            name=name, old=oldAttr.value, new=value,
                            new_file=filestack[-1]))
            else:
                parent.setAttribute(name, value)


def process_include(elt, macros, symbols, func):
    included = []
    filename_spec, namespace_spec, optional = check_attrs(
        elt, ['filename'], ['ns', 'optional'])
    if namespace_spec:
        try:
            namespace_spec = eval_text(namespace_spec, symbols)
            macros[namespace_spec] = ns_macros = MacroNameSpace()
            symbols[namespace_spec] = ns_symbols = PropertyNameSpace()
            macros = ns_macros
            symbols = ns_symbols
        except TypeError:   # pragma: no cover
            raise XacroException(
                'namespaces are supported with in-order option only')

    optional = get_boolean_value(optional, None)

    if first_child_element(elt):
        warning("Child elements of a <xacro:include> tag are ignored")   # pragma: no cover # noqa
        if verbosity > 0:   # pragma: no cover
            print_location(filestack)

    for filename in get_include_files(filename_spec, symbols):
        try:
            # extend filestack
            oldstack = push_file(filename)
            include = parse(None, filename).documentElement

            # recursive call to func
            func(include, macros, symbols)
            included.append(include)
            import_xml_namespaces(elt.parentNode, include.attributes)
        except XacroException as e:
            if e.exc and isinstance(e.exc, IOError) and optional is True:
                continue
            else:
                raise
        finally:
            # restore filestack
            restore_filestack(oldstack)

    remove_previous_comments(elt)
    # replace the include tag with the nodes of the included file(s)
    replace_node(elt, by=included, content_only=True)


def is_valid_name(name):
    """
    Checks whether name is a valid property or macro identifier.
    With python-based evaluation, we need to avoid name clashes with python
    keywords.
    """
    # Resulting AST of simple identifier is <Module [<Expr <Name "foo">>]>
    try:
        root = ast.parse(name)

        if isinstance(root, ast.Module) and \
           len(root.body) == 1 and isinstance(root.body[0], ast.Expr) and \
           isinstance(
               root.body[0].value, ast.Name) and root.body[0].value.id == name:
            return True
    except SyntaxError:
        pass

    return False


re_macro_arg = re.compile(r'''\s*([^\s:=]+?):?=(\^\|?)?((?:(?:'[^']*')?[^\s'"]*?)*)(?:\s+|$)(.*)''')    # noqa
#                           space   param    :=   ^|   <--      default      -->   space    rest        # noqa


def parse_macro_arg(s):
    """
    parse the first param spec from a macro parameter string s
    accepting the following syntax: <param>[:=|=][^|]<default>
    :param s: param spec string
    :return: param, (forward, default), rest-of-string
             forward will be either param or None (depending on whether
             ^ was specified)
             default will be the default string or None
             If there is no default spec at all, the middle pair will be
             replaced by None
    """
    m = re_macro_arg.match(s)
    if m:
        # there is a default value specified for param
        param, forward, default, rest = m.groups()
        if not default:
            default = None
        return param, (param if forward else None, default), rest
    else:
        # there is no default specified at all
        result = s.lstrip().split(None, 1)
        return result[0], None, result[1] if len(result) > 1 else ''


def grab_macro(elt, macros):
    assert(elt.tagName == 'xacro:macro')
    remove_previous_comments(elt)

    name, params = check_attrs(elt, ['name'], ['params'])
    if name == 'call':
        raise XacroException("Invalid use of macro name 'call'")
    if name.find('.') != -1:   # pragma: no cover
        raise XacroException(
            "macro names must not contain '.' (reserved for "
            "namespaces): %s" % name)
    if name.startswith('xacro:'):
        warning("macro names must not contain prefix 'xacro:': %s" % name)
        name = name[6:]  # drop 'xacro:' prefix

    # fetch existing or create new macro definition
    macro = macros.get(name, Macro())
    # append current filestack to history
    macro.history.append(filestack)
    macro.body = elt

    # parse params and their defaults
    macro.params = []
    macro.defaultmap = {}
    while params:
        param, value, params = parse_macro_arg(params)
        macro.params.append(param)
        if value is not None:
            macro.defaultmap[param] = value  # parameter with default

    macros[name] = macro
    replace_node(elt, by=None)


def grab_property(elt, table):
    assert(elt.tagName == 'xacro:property')
    remove_previous_comments(elt)

    name, value, default, scope = check_attrs(
        elt, ['name'], ['value', 'default', 'scope'])
    if not is_valid_name(name):
        raise XacroException(
            'Property names must be valid python identifiers: ' + name)
    if value is not None and default is not None:   # pragma: no cover
        raise XacroException(
            'Property cannot define both a default and a value: ' + name)

    if default is not None:
        if scope is not None:   # pragma: no cover
            warning(
                "%s: default property value can only be defined "
                "on local scope" % name)
        if name not in table:
            value = default
        else:   # pragma: no cover
            replace_node(elt, by=None)
            return

    if value is None:
        name = '**' + name
        value = elt  # debug

    replace_node(elt, by=None)

    if scope and scope == 'global':
        target_table = table.root()
        unevaluated = False
    elif scope and scope == 'parent':
        if table.parent:
            target_table = table.parent
            unevaluated = False
        else:   # pragma: no cover
            warning("%s: no parent scope at global scope " % name)
            return  # cannot store the value, no reason to evaluate it
    else:
        target_table = table
        unevaluated = True

    if not unevaluated and isinstance(value, _basestr):
        value = eval_text(value, table)

    target_table._setitem(name, value, unevaluated=unevaluated)


LEXER = QuickLexer(
    # multiple $ in a row, followed by { or (
    DOLLAR_DOLLAR_BRACE=r"^\$\$+(\{|\()",
    EXPR=r"^\$\{[^\}]*\}",        # stuff starting with ${
    EXTENSION=r"^\$\([^\)]*\)",   # stuff starting with $(
    # any text w/o $ or  $ following any chars except {($  or  single $
    TEXT=r"[^$]+|\$[^{($]+|\$$")


# evaluate text and return typed value
def eval_text(text, symbols):
    def handle_expr(s):
        try:
            return eval(eval_text(s, symbols), global_symbols, symbols)
        except Exception as e:
            # re-raise as XacroException to add more context
            raise XacroException(exc=e,
                                 suffix=os.linesep
                                 + "when evaluating expression '%s'" % s)

    def handle_extension(s):   # pragma: no cover
        return eval_extension("$(%s)" % eval_text(s, symbols))

    results = []
    lex = QuickLexer(LEXER)
    lex.lex(text)
    while lex.peek():
        id = lex.peek()[0]
        if id == lex.EXPR:
            results.append(handle_expr(lex.next()[1][2:-1]))
        elif id == lex.EXTENSION:   # pragma: no cover
            results.append(handle_extension(lex.next()[1][2:-1]))
        elif id == lex.TEXT:
            results.append(lex.next()[1])
        elif id == lex.DOLLAR_DOLLAR_BRACE:
            results.append(lex.next()[1][1:])
    # return single element as is, i.e. typed
    if len(results) == 1:
        return results[0]
    # otherwise join elements to a string
    else:
        return ''.join(map(unicode, results))


def eval_default_arg(forward_variable, default, symbols, macro):
    if forward_variable is None:
        return eval_text(default, symbols)
    try:
        return symbols[forward_variable]
    except KeyError:   # pragma: no cover
        if default is not None:
            return eval_text(default, symbols)
        else:
            raise XacroException(
                "Undefined property to forward: "
                + forward_variable, macro=macro)


def handle_dynamic_macro_call(node, macros, symbols):
    name, = reqd_attrs(node, ['macro'])
    if not name:   # pragma: no cover
        raise XacroException("xacro:call is missing the 'macro' attribute")
    name = unicode(eval_text(name, symbols))

    # remove 'macro' attribute and rename tag with resolved macro name
    node.removeAttribute('macro')
    node.tagName = 'xacro:' + name
    # forward to handle_macro_call
    handle_macro_call(node, macros, symbols)
    return True


def resolve_macro(fullname, macros):
    # split name into namespaces and real name
    namespaces = fullname.split('.')
    name = namespaces.pop(-1)

    def _resolve(namespaces, name, macros):
        # traverse namespaces to actual macros dict
        for ns in namespaces:
            macros = macros[ns]
        return macros[name]

    # try fullname and (namespaces, name) in this order
    try:
        return _resolve([], fullname, macros)
    except KeyError:
        if namespaces:
            return _resolve(namespaces, name, macros)
        else:
            raise


def handle_macro_call(node, macros, symbols):
    if node.tagName == 'xacro:call':
        return handle_dynamic_macro_call(node, macros, symbols)
    elif not node.tagName.startswith('xacro:'):
        return False  # no macro

    name = node.tagName[6:]  # drop 'xacro:' prefix
    try:
        m = resolve_macro(name, macros)
        body = m.body.cloneNode(deep=True)

    except KeyError:
        raise XacroException("unknown macro name: %s" % node.tagName)

    # Expand the macro
    scoped = Table(symbols)  # new local name space for macro evaluation
    params = m.params[:]  # deep copy macro's params list
    for name, value in node.attributes.items():
        if name not in params:   # pragma: no cover
            raise XacroException(
                "Invalid parameter \"%s\"" % unicode(name), macro=m)
        params.remove(name)
        scoped._setitem(name, eval_text(value, symbols), unevaluated=False)
        node.setAttribute(name, "")  # suppress second evaluation in eval_all()

    # Evaluate block parameters in node
    eval_all(node, macros, symbols)

    # Fetch block parameters, in order
    block = first_child_element(node)
    for param in params[:]:
        if param[0] == '*':
            if not block:   # pragma: no cover
                raise XacroException("Not enough blocks", macro=m)
            params.remove(param)
            scoped[param] = block
            block = next_sibling_element(block)

    if block is not None:   # pragma: no cover
        raise XacroException("Unused block \"%s\"" % block.tagName, macro=m)

    # Try to load defaults for any remaining non-block parameters
    for param in params[:]:
        # block parameters are not supported for defaults
        if param[0] == '*':   # pragma: no cover
            continue

        # get default
        name, default = m.defaultmap.get(param, (None, None))
        if name is not None or default is not None:
            scoped._setitem(
                param, eval_default_arg(
                    name, default, symbols, m), unevaluated=False)
            params.remove(param)

    if params:
        raise XacroException(
            "Undefined parameters [%s]" % ",".join(params), macro=m)

    try:
        eval_all(body, macros, scoped)
    except Exception as e:   # pragma: no cover
        # fill in macro call history for nice error reporting
        if hasattr(e, 'macros'):
            e.macros.append(m)
        else:
            e.macros = [m]
        raise

    # Replaces the macro node with the expansion
    remove_previous_comments(node)
    replace_node(node, by=body, content_only=True)
    return True


def get_boolean_value(value, condition):
    """
    Return a boolean value that corresponds to the given Xacro condition value.
    Values "true", "1" and "1.0" are supposed to be True.
    Values "false", "0" and "0.0" are supposed to be False.
    All other values raise an exception.
    :param value: The value to be evaluated. The value has to already be
        evaluated by Xacro.
    :param condition: The original condition text in the XML.
    :return: The corresponding boolean value, or a Python expression that,
        converted to boolean, corresponds to it.
    :raises ValueError: If the condition value is incorrect.
    """
    try:
        if isinstance(value, _basestr):
            if value == 'true' or value == 'True':
                return True
            elif value == 'false' or value == 'False':
                return False
            else:
                return bool(int(value))
        else:
            return bool(value)
    except Exception:
        raise XacroException(
            "Xacro conditional \"%s\" evaluated to \"%s\", "
            "which is not a boolean expression." % (condition, value))


_empty_text_node = xml.dom.minidom.getDOMImplementation().createDocument(
    None, "dummy", None).createTextNode('\n\n')


def remove_previous_comments(node):
    """remove consecutive comments in front of the xacro-specific node"""
    next = node.nextSibling
    previous = node.previousSibling
    while previous:
        if previous.nodeType == xml.dom.Node.TEXT_NODE and \
                previous.data.isspace() and previous.data.count('\n') <= 1:
            # skip a single empty text node (max 1 newline)
            previous = previous.previousSibling

        if previous and previous.nodeType == xml.dom.Node.COMMENT_NODE:
            comment = previous
            previous = previous.previousSibling
            node.parentNode.removeChild(comment)
        else:
            # insert empty text node to stop removing of comments in future
            # calls actually this moves the singleton instance to the new
            # location
            if next and _empty_text_node != next:
                node.parentNode.insertBefore(_empty_text_node, next)
            return


def eval_all(node, macros, symbols):
    """
    Recursively evaluate node, expanding macros, replacing properties,
    and evaluating expressions
    """
    # evaluate the attributes
    for name, value in node.attributes.items():
        if name.startswith('xacro:'):  # remove xacro:* attributes
            node.removeAttribute(name)
        else:
            result = unicode(eval_text(value, symbols))
            node.setAttribute(name, result)

    # remove xacro namespace definition
    try:
        node.removeAttribute('xmlns:xacro')
    except xml.dom.NotFoundErr:
        pass

    node = node.firstChild
    while node:
        next = node.nextSibling
        if node.nodeType == xml.dom.Node.ELEMENT_NODE:
            if node.tagName == 'xacro:insert_block':
                name, = check_attrs(node, ['name'], [])

                if ("**" + name) in symbols:
                    # Multi-block
                    block = symbols['**' + name]
                    content_only = True
                elif ("*" + name) in symbols:
                    # Single block
                    block = symbols['*' + name]
                    content_only = False
                else:   # pragma: no cover
                    raise XacroException("Undefined block \"%s\"" % name)

                # cloning block allows to insert the same block multiple times
                block = block.cloneNode(deep=True)
                # recursively evaluate block
                eval_all(block, macros, symbols)
                replace_node(node, by=block, content_only=content_only)

            elif node.tagName == 'xacro:include':
                process_include(node, macros, symbols, eval_all)

            elif node.tagName == 'xacro:property':
                grab_property(node, symbols)

            elif node.tagName == 'xacro:macro':
                grab_macro(node, macros)

            elif node.tagName == 'xacro:arg':
                name, default = check_attrs(node, ['name', 'default'], [])
                if name not in substitution_args_context['arg']:
                    substitution_args_context['arg'][name] = eval_text(
                        default, symbols)

                remove_previous_comments(node)
                replace_node(node, by=None)

            elif node.tagName == 'xacro:element':
                name = eval_text(
                    *reqd_attrs(node, ['xacro:name']), symbols=symbols)
                if not name:   # pragma: no cover
                    raise XacroException("xacro:element: empty name")

                node.removeAttribute('xacro:name')
                node.nodeName = node.tagName = name
                continue  # re-process the node with new tagName

            elif node.tagName == 'xacro:attribute':
                name, value = [eval_text(a, symbols) for a in reqd_attrs(
                    node, ['name', 'value'])]
                if not name:   # pragma: no cover
                    raise XacroException("xacro:attribute: empty name")

                node.parentNode.setAttribute(name, value)
                replace_node(node, by=None)

            elif node.tagName in ['xacro:if', 'xacro:unless']:
                remove_previous_comments(node)
                cond, = check_attrs(node, ['value'], [])
                keep = get_boolean_value(eval_text(cond, symbols), cond)
                if node.tagName in ['unless', 'xacro:unless']:
                    keep = not keep

                if keep:
                    eval_all(node, macros, symbols)
                    replace_node(node, by=node, content_only=True)
                else:
                    replace_node(node, by=None)

            elif handle_macro_call(node, macros, symbols):
                # handle_macro_call does all the work of expanding the macro
                pass

            else:
                eval_all(node, macros, symbols)

        # TODO: Also evaluate content of COMMENT_NODEs?
        elif node.nodeType == xml.dom.Node.TEXT_NODE:
            node.data = unicode(eval_text(node.data, symbols))

        node = next


def parse(inp, filename=None):
    """
    Parse input or filename into a DOM tree.
    If inp is None, open filename and load from there.
    Otherwise, parse inp, either as string or file object.
    If inp is already a DOM tree, this function is a noop.
    :return:xml.dom.minidom.Document
    :raise: xml.parsers.expat.ExpatError
    """

    f = None
    if inp is None:
        try:
            inp = f = open(filename)
        except IOError as e:
            # do not report currently processed file as "in file ..."
            filestack.pop()
            raise XacroException(e.strerror + ": " + e.filename, exc=e)

    try:
        if isinstance(inp, _basestr):
            return xml.dom.minidom.parseString(inp)
        elif hasattr(inp, 'read'):
            return xml.dom.minidom.parse(inp)
        return inp   # pragma: no cover

    finally:
        if f:
            f.close()


def process_doc(doc, mappings=None, **kwargs):
    global verbosity
    verbosity = kwargs.get('verbosity', verbosity)

    # set substitution args
    substitution_args_context['arg'] = {} if mappings is None else mappings

    # if not yet defined: initialize filestack
    if not filestack:
        restore_filestack([None])

    macros = {}
    symbols = Table()

    # apply xacro:targetNamespace as global xmlns (if defined)
    targetNS = doc.documentElement.getAttribute('xacro:targetNamespace')
    if targetNS:
        doc.documentElement.removeAttribute('xacro:targetNamespace')
        doc.documentElement.setAttribute('xmlns', targetNS)

    eval_all(doc.documentElement, macros, symbols)

    # reset substitution args
    substitution_args_context['arg'] = {}


def open_output(output_filename):
    if output_filename is None:   # pragma: no cover
        return sys.stdout
    else:   # pragma: no cover
        dir_name = os.path.dirname(output_filename)
        if dir_name:
            try:
                os.makedirs(dir_name)
            except os.error:
                # errors occur when dir_name exists or creation failed
                # ignore error here; opening of file will fail if directory
                # is still missing
                pass

        try:
            return open(output_filename, 'w')
        except IOError as e:
            raise XacroException("Failed to open output:", exc=e)


def print_location(filestack, err=None, file=sys.stderr):
    macros = getattr(err, 'macros', []) if err else []
    msg = 'when instantiating macro:'
    for m in macros:   # pragma: no cover
        name = m.body.getAttribute('name')
        location = '(%s)' % m.history[-1][-1]
        print(msg, name, location, file=file)
        msg = 'instantiated from:'

    msg = 'in file:' if macros else 'when processing file:'
    for f in reversed(filestack):
        if f is None:
            f = 'string'
        print(msg, f, file=file)
        msg = 'included from:'


def process_file(input_file_name, **kwargs):   # pragma: no cover
    """main processing pipeline"""
    # initialize file stack for error-reporting
    restore_filestack([input_file_name])
    # parse the document into a xml.dom tree
    doc = parse(None, input_file_name)
    # perform macro replacement
    process_doc(doc, **kwargs)

    # add xacro auto-generated banner
    banner = [xml.dom.minidom.Comment(c) for c in
              [" %s " % ('=' * 83),
               " |    This document was autogenerated by xacro from %-30s | "
               % input_file_name,
               " |    EDITING THIS FILE BY HAND IS NOT RECOMMENDED  %-30s | "
               % "",
               " %s " % ('=' * 83)]]
    first = doc.firstChild
    for comment in banner:
        doc.insertBefore(comment, first)

    return doc


def main(filename, tld_other=None):   # pragma: no cover
    opts = {
        'output': None,
        'just_deps': False,
        'in_order': True,
        'verbosity': 1,
        'just_includes': False,
        'mappings': {}
    }

    global tld

    if tld_other is None:
        tld = '..'
    else:
        tld = tld_other

    try:
        # open and process file
        doc = process_file(filename, **opts)
        # open the output file
        out = open_output(opts["output"])

    except Exception as e:
        msg = unicode(e)
        if not msg:
            msg = repr(e)
        error(msg)
        if verbosity > 0:
            print_location(filestack, e)
        if verbosity > 1:
            print(file=sys.stderr)  # add empty separator line before error
            raise  # create stack trace
        else:
            sys.exit(2)  # gracefully exit with error condition

    # special output mode
    if opts["just_deps"]:
        out.write(" ".join(set(all_includes)))
        print()
        return

    # write output
    # out.write(doc.toprettyxml(indent='  ', **encoding))
    # print()
    # only close output file, but not stdout
    if opts["output"]:
        out.close()

    return doc.toprettyxml(indent='  ', **encoding)



#!/usr/bin/env python3
"""
Created on Tue Apr 24 15:48:52 2020
@author: Jesse Haviland
"""

import sys
import os
from os.path import splitext
import tempfile
import subprocess
import webbrowser
import numpy as np
# import spatialmath as sp
from spatialmath import SE3
from spatialmath.base.argcheck import getvector, verifymatrix, getmatrix
from roboticstoolbox.robot.ELink import ELink, ETS
# from roboticstoolbox.backends.PyPlot.functions import \
#     _plot, _teach, _fellipse, _vellipse, _plot_ellipse, \
#     _plot2, _teach2
from roboticstoolbox.tools import xacro
from roboticstoolbox.tools import URDF
from roboticstoolbox.robot.Robot import Robot
from roboticstoolbox.robot.Gripper import Gripper
from pathlib import PurePath, PurePosixPath
from ansitable import ANSITable, Column
from spatialmath import SpatialAcceleration, SpatialVelocity, \
    SpatialInertia, SpatialForce


class ERobot(Robot):
    """
    The ERobot. A superclass which represents the
    kinematics of a serial-link manipulator

    :param et_list: List of elementary transforms which represent the robot
        kinematics
    :type et_list: ET list
    :param name: Name of the robot
    :type name: str, optional
    :param manufacturer: Manufacturer of the robot
    :type manufacturer: str, optional
    :param base: Location of the base is the world frame
    :type base: SE3, optional
    :param tool: Offset of the flange of the robot to the end-effector
    :type tool: SE3, optional
    :param gravity: The gravity vector
    :type n: ndarray(3)

    :references:
        - Kinematic Derivatives using the Elementary Transform Sequence,
          J. Haviland and P. Corke
    """

    # TODO do we need tool and base as well?

    def __init__(
            self,
            elinks,
            base_link=None,
            gripper_links=None,
            checkjindex=True,
            **kwargs
            ):

        self._ets = []
        self._linkdict = {}
        self._n = 0
        self._ee_links = []
        self._base_link = None

        # Ordered links, we reorder the input elinks to be in depth first
        # search order
        orlinks = []

        link_number = 0
        if isinstance(elinks, ETS):
            # were passed an ETS string
            ets = elinks
            elinks = []

            # chop it up into segments, a link frame after every joint
            start = 0
            for j, k in enumerate(ets.joints()):
                ets_j = ets[start:k+1]
                start = k + 1
                if j == 0:
                    parent = None
                else:
                    parent = elinks[-1]
                elink = ELink(ets_j, parent=parent, name=f"link{j:d}")
                elinks.append(elink)

            n = len(ets.joints())

            tool = ets[start:]
            if len(tool) > 0:
                elinks.append(ELink(tool, parent=elinks[-1], name="ee"))
        elif isinstance(elinks, list):
            # we're passed a list of ELinks

            # check all the incoming ELink objects
            n = 0
            for link in elinks:
                if isinstance(link, ELink):
                    # if link has no name, give it one
                    if link.name is None:
                        link.name = f"link-{link_number}"
                        link_number += 1

                    # put it in the link dictionary, check for duplicates
                    if link.name in self._linkdict:
                        raise ValueError(
                            f'link name {link.name} is not unique')
                    self._linkdict[link.name] = link
                else:
                    raise TypeError("Input can be only ELink")
                if link.isjoint:
                    n += 1
        else:
            raise TypeError('elinks must be a list of ELinks or an ETS')

        self._n = n

        # scan for base
        for link in elinks:
            # is this a base link?
            if link._parent is None:
                if self._base_link is not None:
                    raise ValueError('Multiple base links')
                self._base_link = link
            else:
                # no, update children of this link's parent
                link._parent._child.append(link)

        # Set up the gripper, make a list containing the root of all
        # grippers
        if gripper_links is not None:
            if isinstance(gripper_links, ELink):
                gripper_links = [gripper_links]
        else:
            gripper_links = []

        # An empty list to hold all grippers
        self.grippers = []

        # Make a gripper object for each gripper
        for link in gripper_links:
            g_links = self.dfs_links(link)

            # Remove gripper links from the robot
            for g_link in g_links:
                elinks.remove(g_link)

            # Save the gripper object
            self.grippers.append(Gripper(g_links))

        # Subtract the n of the grippers from the n of the robot
        for gripper in self.grippers:
            self._n -= gripper.n

        # Set the ee links
        self.ee_links = []
        if len(gripper_links) == 0:
            for link in elinks:
                # is this a leaf node? and do we not have any grippers
                if len(link.child) == 0:
                    # no children, must be an end-effector
                    self.ee_links.append(link)
        else:
            for link in gripper_links:
                # use the passed in value
                self.ee_links.append(link.parent)

        # assign the joint indices
        if all([link.jindex is None for link in elinks]):

            jindex = [0]  # "mutable integer" hack

            def visit_link(link, jindex):
                # if it's a joint, assign it a jindex and increment it
                if link.isjoint and link in elinks:
                    link.jindex = jindex[0]
                    jindex[0] += 1

                if link in elinks:
                    orlinks.append(link)

            # visit all links in DFS order
            self.dfs_links(
                self.base_link, lambda link: visit_link(link, jindex))

        elif all([link.jindex is not None for link in elinks if link.isjoint]):
            # jindex set on all, check they are unique and sequential
            if checkjindex:
                jset = set(range(self._n))
                for link in elinks:
                    if link.isjoint and link.jindex not in jset:
                        raise ValueError(
                            f'joint index {link.jindex} was '
                            'repeated or out of range')
                    jset -= set([link.jindex])
                if len(jset) > 0:  # pragma nocover  # is impossible
                    raise ValueError(f'joints {jset} were not assigned')
            orlinks = elinks
        else:
            # must be a mixture of ELinks with/without jindex
            raise ValueError(
                'all links must have a jindex, or none have a jindex')

        # Current joint angles of the robot
        # TODO should go to Robot class?
        self.q = np.zeros(self.n)
        self.qd = np.zeros(self.n)
        self.qdd = np.zeros(self.n)
        self.control_type = 'v'

        super().__init__(orlinks, **kwargs)

    def dfs_links(self, start, func=None):
        """
        Visit all links from start in depth-first order and will apply
        func to each visited link

        :param start: the link to start at
        :type start: ELink
        :param func: An optional function to apply to each link as it is found
        :type func: function

        :returns: A list of links
        :rtype: list of ELink
        """
        visited = []

        def vis_children(link):
            visited.append(link)
            if func is not None:
                func(link)

            for li in link.child:
                if li not in visited:
                    vis_children(li)

        vis_children(start)

        return visited

    # def dfs_path(self, l1, l2):
    #     path = []
    #     visited = [l1]

    #     def vis_children(link):
    #         visited.append(link)

    #         for li in link.child:
    #             if li not in visited:

    #                 if li == l2 or vis_children(li):
    #                     path.append(li)
    #                     return True
    #     vis_children(l1)
    #     path.append(l1)
    #     path.reverse()
    #     return path

    def to_dict(self):
        ob = {
            'links': [],
            'name': self.name,
            'n': self.n
        }

        self.fkine_all(self.q)

        for link in self.links:
            li = {
                'axis': [],
                'eta': [],
                'geometry': [],
                'collision': []
            }

            for et in link.ets():
                li['axis'].append(et.axis)
                li['eta'].append(et.eta)

            if link.v is not None:
                li['axis'].append(link.v.axis)
                li['eta'].append(link.v.eta)

            for gi in link.geometry:
                li['geometry'].append(gi.to_dict())

            for gi in link.collision:
                li['collision'].append(gi.to_dict())

            ob['links'].append(li)

        # Do the grippers now
        for gripper in self.grippers:
            for link in gripper.links:
                li = {
                    'axis': [],
                    'eta': [],
                    'geometry': [],
                    'collision': []
                }

                for et in link.ets():
                    li['axis'].append(et.axis)
                    li['eta'].append(et.eta)

                if link.v is not None:
                    li['axis'].append(link.v.axis)
                    li['eta'].append(link.v.eta)

                for gi in link.geometry:
                    li['geometry'].append(gi.to_dict())

                for gi in link.collision:
                    li['collision'].append(gi.to_dict())

                ob['links'].append(li)

        return ob

    def fk_dict(self):
        ob = {
            'links': []
        }

        self.fkine_all(self.q)

        # Do the robot
        for link in self.links:

            li = {
                'geometry': [],
                'collision': []
            }

            for gi in link.geometry:
                li['geometry'].append(gi.fk_dict())

            for gi in link.collision:
                li['collision'].append(gi.fk_dict())

            ob['links'].append(li)

        # Do the grippers now
        for gripper in self.grippers:
            for link in gripper.links:
                li = {
                    'geometry': [],
                    'collision': []
                }

                for gi in link.geometry:
                    li['geometry'].append(gi.fk_dict())

                for gi in link.collision:
                    li['collision'].append(gi.fk_dict())

                ob['links'].append(li)

        return ob

    # @classmethod
    # def urdf_to_ets(cls, file_path):
    #     name, ext = splitext(file_path)

    #     if ext == '.xacro':
    #         urdf_string = xacro.main(file_path)
    #         urdf = URDF.loadstr(urdf_string, file_path)

    #     return ERobot(
    #         urdf.elinks,
    #         name=urdf.name
    #     )

    def urdf_to_ets_args(self, file_path, tld=None):
        """
        [summary]

        :param file_path: File path relative to the xacro folder
        :type file_path: str, in Posix file path fprmat
        :param tld: top-level directory, defaults to None
        :type tld: str, optional
        :return: Links and robot name
        :rtype: tuple(ELink list, str)
        """

        # get the path to the class that defines the robot
        classpath = sys.modules[self.__module__].__file__
        # print("*** urdf_to_ets_args: ", classpath)
        # add on relative path to get to the URDF or xacro file
        base_path = PurePath(classpath).parent.parent / 'URDF' / 'xacro'
        file_path = base_path / PurePosixPath(file_path)
        name, ext = splitext(file_path)

        if ext == '.xacro':
            # it's a xacro file, preprocess it
            if tld is not None:
                tld = base_path / PurePosixPath(tld)
            urdf_string = xacro.main(file_path, tld)
            urdf = URDF.loadstr(urdf_string, file_path)
        else:  # pragma nocover
            urdf = URDF.loadstr(open(file_path).read(), file_path)

        return urdf.elinks, urdf.name

    # @classmethod
    # def dh_to_ets(cls, robot):
    #     """
    #     Converts a robot modelled with standard or modified DH parameters to
    #     an ERobot representation

    #     :param robot: The robot model to be converted
    #     :type robot: SerialLink
    #     :return: List of returned :class:`bluepy.btle.Characteristic` objects
    #     :rtype: ets class
    #     """
    #     ets = []
    #     q_idx = []
    #     M = 0

    #     for j in range(robot.n):
    #         L = robot.links[j]

    #         # Method for modified DH parameters
    #         if robot.mdh:

    #             # Append Tx(a)
    #             if L.a != 0:
    #                 ets.append(ET.Ttx(L.a))
    #                 M += 1

    #             # Append Rx(alpha)
    #             if L.alpha != 0:
    #                 ets.append(ET.TRx(L.alpha))
    #                 M += 1

    #             if L.is_revolute:
    #                 # Append Tz(d)
    #                 if L.d != 0:
    #                     ets.append(ET.Ttz(L.d))
    #                     M += 1

    #                 # Append Rz(q)
    #                 ets.append(ET.TRz(joint=j+1))
    #                 q_idx.append(M)
    #                 M += 1

    #             else:
    #                 # Append Tz(q)
    #                 ets.append(ET.Ttz(joint=j+1))
    #                 q_idx.append(M)
    #                 M += 1

    #                 # Append Rz(theta)
    #                 if L.theta != 0:
    #                     ets.append(ET.TRz(L.alpha))
    #                     M += 1

    #     return cls(
    #         ets,
    #         q_idx,
    #         robot.name,
    #         robot.manuf,
    #         robot.base,
    #         robot.tool)

    @property
    def qlim(self):
        v = np.zeros((2, self.n))
        j = 0

        for i in range(len(self.links)):
            if self.links[i].isjoint:
                v[:, j] = self.links[i].qlim
                j += 1

        return v

    # @property
    # def qdlim(self):
    #     return self.qdlim

# --------------------------------------------------------------------- #

    @property
    def n(self):
        return self._n
# --------------------------------------------------------------------- #

    @property
    def elinks(self):
        # return self._linkdict
        return self._links
# --------------------------------------------------------------------- #

    @property
    def link_dict(self):
        return self._linkdict
# --------------------------------------------------------------------- #

    @property
    def base_link(self):
        return self._base_link

    @base_link.setter
    def base_link(self, link):
        if isinstance(link, ELink):
            self._base_link = link
        else:
            # self._base_link = self.links[link]
            raise TypeError('Must be an ELink')
        # self._reset_fk_path()
# --------------------------------------------------------------------- #
    # TODO  get configuration string

    @property
    def ee_links(self):
        return self._ee_links

    # def add_ee(self, link):
    #     if isinstance(link, ELink):
    #         self._ee_link.append(link)
    #     else:
    #         raise ValueError('must be an ELink')
    #     self._reset_fk_path()

    @ee_links.setter
    def ee_links(self, link):
        if isinstance(link, ELink):
            self._ee_links = [link]
        elif isinstance(link, list) and \
                all([isinstance(x, ELink) for x in link]):
            self._ee_links = link
        else:
            raise TypeError('expecting an ELink or list of ELinks')
        # self._reset_fk_path()

    @property
    def reach(self):
        r"""
        Reach of the robot

        :return: Maximum reach of the robot
        :rtype: float

        A conservative estimate of the reach of the robot. It is computed as
        the sum of the translational ETs that define the link transform.

        .. note::

            - Probably an overestimate of reach
            - Used by numerical inverse kinematics to scale translational
              error.
            - For a prismatic joint, uses ``qlim`` if it is set

        .. warning:: Computed on the first access. If kinematic parameters
              subsequently change this will not be reflected.
        """
        if self._reach is None:
            d = 0
            for link in self:
                for et in link.ets():
                    if et.isprismatic:
                        d += abs(et.eta)
                if link.isprismatic and link.qlim is not None:
                    d += link.qlim[1]
            self._reach = d
        return self._reach

# --------------------------------------------------------------------- #

    # @property
    # def ets(self):
    #     return self._ets

# --------------------------------------------------------------------- #

    # @property
    # def M(self):
    #     return self._M

# --------------------------------------------------------------------- #

    # @property
    # def q_idx(self):
    #     return self._q_idx

# --------------------------------------------------------------------- #

    # def ets(self, ee=None):
    #     if ee is None:
    #         if len(self.ee_links) == 1:
    #             link = self.ee_links[0]
    #         else:
    #             raise ValueError(
    #                 'robot has multiple end-effectors, specify one')
    #     # elif isinstance(ee, str) and ee in self._linkdict:
    #     #     ee = self._linkdict[ee]
    #     elif isinstance(ee, ELink) and ee in self._links:
    #         link = ee
    #     else:
    #         raise ValueError('end-effector is not valid')

    #     ets = ETS()

    #     # build the ETS string from ee back to root
    #     while link is not None:
    #         ets = link.ets() * ets
    #         link = link.parent

    #     return ets

    def ets(self, start=None, end=None, explored=None, path=None):
        """
        ERobot to ETS

        :param start: start of path, defaults to ``base_link``
        :type start: ELink or str, optional
        :param end: end of path, defaults to end-effector
        :type end: ELink or str, optional
        :raises ValueError: a link does not belong to this ERobot
        :raises TypeError: a bad link argument
        :return: elementary transform sequence
        :rtype: ETS instance

        - ``robot.ets()`` is an ETS representing the kinematics from base to
          end-effector.
        - ``robot.ets(end=link)`` is an ETS representing the kinematics from
          base to the link ``link`` specified as an ELink reference or a name.
        - ``robot.ets(start=l1, end=l2)`` is an ETS representing the kinematics
          from link ``l1`` to link ``l2``.
        """
        v = self._getlink(start, self.base_link)
        if end is None and len(self.ee_links) > 1:
            raise ValueError(
                'ambiguous, specify which end-effector is required')
        end = self._getlink(end, self.ee_links[0])

        if explored is None:
            explored = set()
        toplevel = path is None

        explored.add(v)
        if v == end:
            return path

        # unlike regular DFS, the neighbours of the node are its children
        # and its parent.

        # visit child nodes
        if toplevel:
            path = v.ets()
        for w in v.child:
            if w not in explored:
                p = self.ets(w, end, explored, path * w.ets())
                if p:
                    return p

        # visit parent node
        if toplevel:
            path = ETS()
        if v.parent is not None:
            w = v.parent
            if w not in explored:
                p = self.ets(w, end, explored, path * v.ets().inv())
                if p:
                    return p

        return None

    def config(self):
        s = ''
        for link in self.links:
            if link.v is not None:
                if link.v.isprismatic:
                    s += 'P'
                elif link.v.isrevolute:
                    s += 'R'
        return s

# --------------------------------------------------------------------- #

    def showgraph(self, **kwargs):
        """
        Display a link transform graph in browser

        :param etsbox: Put the link ETS in a box, otherwise an edge label
        :type etsbox: bool
        :param jtype: Arrowhead to node indicates revolute or prismatic type
        :type jtype: bool
        :param static: Show static joints in blue and bold
        :type static: bool

        ``robot.showgraph()`` displays a graph of the robot's link frames
        and the ETS between them.  It uses GraphViz dot.

        The nodes are:

            - Base is shown as a grey square.  This is the world frame origin,
              but can be changed using the ``base`` attribute of the robot.
            - Link frames are indicated by circles
            - ETS transforms are indicated by rounded boxes

        The edges are:

            - an arrow if `jtype` is False or the joint is fixed
            - an arrow with a round head if `jtype` is True and the joint is
              revolute
            - an arrow with a box head if `jtype` is True and the joint is
              prismatic

        Edge labels or nodes in blue have a fixed transformation to the
        preceding link.

        :seealso: :func:`dotfile`
        """

        # create the temporary dotfile
        dotfile = tempfile.TemporaryFile(mode="w")
        self.dotfile(dotfile, **kwargs)

        # rewind the dot file, create PDF file in the filesystem, run dot
        dotfile.seek(0)
        pdffile = tempfile.NamedTemporaryFile(suffix=".pdf", delete=False)
        subprocess.run("dot -Tpdf", shell=True, stdin=dotfile, stdout=pdffile)

        # open the PDF file in browser (hopefully portable), then cleanup
        webbrowser.open(f"file://{pdffile.name}")
        os.remove(pdffile.name)

    def dotfile(self, filename, etsbox=False, jtype=False, static=True):
        """
        Write a link transform graph as a GraphViz dot file

        :param file: Name of file to write to
        :type file: str or file
        :param etsbox: Put the link ETS in a box, otherwise an edge label
        :type etsbox: bool
        :param jtype: Arrowhead to node indicates revolute or prismatic type
        :type jtype: bool
        :param static: Show static joints in blue and bold
        :type static: bool

        The file can be processed using dot::

            % dot -Tpng -o out.png dotfile.dot

        The nodes are:

            - Base is shown as a grey square.  This is the world frame origin,
              but can be changed using the ``base`` attribute of the robot.
            - Link frames are indicated by circles
            - ETS transforms are indicated by rounded boxes

        The edges are:

            - an arrow if `jtype` is False or the joint is fixed
            - an arrow with a round head if `jtype` is True and the joint is
              revolute
            - an arrow with a box head if `jtype` is True and the joint is
              prismatic

        Edge labels or nodes in blue have a fixed transformation to the
        preceding link.

        .. note:: If ``filename`` is a file object then the file will *not*
            be closed after the GraphViz model is written.

        :seealso: :func:`showgraph`
        """
        if isinstance(filename, str):
            file = open(filename, 'w')
        else:
            file = filename

        header = r"""digraph G {

graph [rankdir=LR];

"""

        def draw_edge(link, etsbox, jtype, static):
            # draw the edge
            if jtype:
                if link.isprismatic:
                    edge_options = \
                        'arrowhead="box", arrowtail="inv", dir="both"'
                elif link.isrevolute:
                    edge_options = \
                        'arrowhead="dot", arrowtail="inv", dir="both"'
                else:
                    edge_options = 'arrowhead="normal"'
            else:
                edge_options = 'arrowhead="normal"'

            if link.parent is None:
                parent = 'BASE'
            else:
                parent = link.parent.name

            if etsbox:
                # put the ets fragment in a box
                if not link.isjoint and static:
                    node_options = ', fontcolor="blue"'
                else:
                    node_options = ''
                file.write(
                    '  {}_ets [shape=box, style=rounded, '
                    'label="{}"{}];\n'.format(
                        link.name, link.ets().__str__(
                            q=f"q{link.jindex}"), node_options))
                file.write('  {} -> {}_ets;\n'.format(parent, link.name))
                file.write('  {}_ets -> {} [{}];\n'.format(
                    link.name, link.name, edge_options))
            else:
                # put the ets fragment as an edge label
                if not link.isjoint and static:
                    edge_options += 'fontcolor="blue"'
                file.write('  {} -> {} [label="{}", {}];\n'.format(
                    parent, link.name, link.ets().__str__(
                        q=f"q{link.jindex}"), edge_options))

        file.write(header)

        # add the base link
        file.write('  BASE [shape=square, style=filled, fillcolor=gray]\n')

        # add the links
        for link in self:
            # draw the link frame node (circle) or ee node (doublecircle)
            if link in self.ee_links:
                # end-effector
                node_options = \
                    'shape="doublecircle", color="blue", fontcolor="blue"'
            else:
                node_options = 'shape="circle"'

            file.write('  {} [{}];\n'.format(link.name, node_options))

            draw_edge(link, etsbox, jtype, static)

        for gripper in self.grippers:
            for link in gripper.links:
                file.write('  {} [shape=cds];\n'.format(link.name))
                draw_edge(link, etsbox, jtype, static)

        file.write('}\n')

        if isinstance(filename, str):
            close(file)  # noqa

# --------------------------------------------------------------------- #

    def fkine(self, q, endlink=None, startlink=None, tool=None):
        '''
        Forward kinematics

        :param q: Joint coordinates
        :type q: ndarray(n) or ndarray(m,n)
        :param endlink: end-effector to compute forward kinematics to
        :type endlink: str or ELink
        :param startlink: the link to compute forward kinematics from
        :type startlink: str or ELink
        :param tool: tool transform, optional
        :type tool: SE3
        :return: The transformation matrix representing the pose of the
            end-effector
        :rtype: SE3 instance

        - ``T = robot.fkine(q)`` evaluates forward kinematics for the robot at
          joint configuration ``q``.

        **Trajectory operation**:

        If ``q`` has multiple rows (mxn), it is considered a trajectory and the
        result is an ``SE3`` instance with ``m`` values.

        .. note::

            - For a robot with a single end-effector there is no need to
              specify ``endlink``
            - For a robot with multiple end-effectors, the ``endlink`` must
              be specified.
            - The robot's base tool transform, if set, is incorporated
              into the result.
            - A tool transform, if provided, is incorporated into the result.
            - Works from the end-effector link to the base

        :references:
            - Kinematic Derivatives using the Elementary Transform
              Sequence, J. Haviland and P. Corke

        '''

        q = getmatrix(q, (None, self.n))

        if tool is not None:
            Ttool = tool.A

        endlink, startlink = self._get_limit_links(endlink, startlink)

        T = SE3.Empty()
        for k, qk in enumerate(q):

            link = endlink  # start with last link

            # add tool if provided
            if tool is None:
                Tk = link.A(qk[link.jindex], fast=True)
            else:
                Tk = link.A(qk[link.jindex], fast=True) @ Ttool

            # add remaining links, back toward the base
            while True:
                link = link.parent

                if link is None:
                    break

                Tk = link.A(qk[link.jindex], fast=True) @ Tk

                if link is startlink:
                    break

            # add base transform if it is set
            if self.base is not None and startlink == self.base_link:
                Tk = self.base.A @ Tk

            T.append(SE3(Tk))

        return T

    def fkine_all(self, q):
        '''
        Tall = robot.fkine_all(q) evaluates fkine for each joint within a
        robot and returns a trajecotry of poses.

        Tall = fkine_all() as above except uses the stored q value of the
        robot object.

        :param q: The joint angles/configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)

        :return T: Homogeneous transformation trajectory
        :rtype T: SE3 list

        .. note::

            - The robot's base transform, if present, are incorporated
              into the result.

        :references:
            - Kinematic Derivatives using the Elementary Transform
              Sequence, J. Haviland and P. Corke

        '''

        q = getvector(q, self.n)

        for link in self.elinks:
            if link.isjoint:
                t = link.A(q[link.jindex])
            else:
                t = link.A()

            # Update the links internal transform wrt the base frame
            if link.parent is None:
                link._fk = self.base * t
            else:
                link._fk = link.parent._fk * t

            # Update the link model transforms as well
            for col in link.collision:
                col.wT = link._fk

            for gi in link.geometry:
                gi.wT = link._fk

        # Do the grippers now
        for gripper in self.grippers:
            for link in gripper.links:
                if link.isjoint:
                    t = link.A(gripper.q[link.jindex])
                else:
                    t = link.A()

                link._fk = link.parent._fk * t

                # Update the link model transforms as well
                for col in link.collision:
                    col.wT = link._fk

                for gi in link.geometry:
                    gi.wT = link._fk

    # def jacob0(self, q=None):
    #     """
    #     J0 = jacob0(q) is the manipulator Jacobian matrix which maps joint
    #     velocity to end-effector spatial velocity. v = J0*qd in the
    #     base frame.

    #     J0 = jacob0() as above except uses the stored q value of the
    #     robot object.

    #     :param q: The joint angles/configuration of the robot (Optional,
    #         if not supplied will use the stored q values).
    #     :type q: float ndarray(n)

    #     :return J: The manipulator Jacobian in ee frame
    #     :rtype: float ndarray(6,n)

    #     :references:
    #         - Kinematic Derivatives using the Elementary Transform
    #           Sequence, J. Haviland and P. Corke
    #     """

    #     if q is None:
    #         q = np.copy(self.q)
    #     else:
    #         q = getvector(q, self.n)

    #     T = (self.base.inv() * self.fkine(q)).A
    #     U = np.eye(4)
    #     j = 0
    #     J = np.zeros((6, self.n))

    #     for link in self._fkpath:

    #         for k in range(link.M):

    #             if k != link.q_idx:
    #                 U = U @ link.ets[k].T().A
    #             else:
    #                 # self._jacoblink(link, k, T)
    #                 U = U @ link.ets[k].T(q[j]).A
    #                 Tu = np.linalg.inv(U) @ T
    #                 n = U[:3, 0]
    #                 o = U[:3, 1]
    #                 a = U[:3, 2]
    #                 x = Tu[0, 3]
    #                 y = Tu[1, 3]
    #                 z = Tu[2, 3]

    #                 if link.ets[k].axis == 'Rz':
    #                     J[:3, j] = (o * x) - (n * y)
    #                     J[3:, j] = a

    #                 elif link.ets[k].axis == 'Ry':
    #                     J[:3, j] = (n * z) - (a * x)
    #                     J[3:, j] = o

    #                 elif link.ets[k].axis == 'Rx':
    #                     J[:3, j] = (a * y) - (o * z)
    #                     J[3:, j] = n

    #                 elif link.ets[k].axis == 'tx':
    #                     J[:3, j] = n
    #                     J[3:, j] = np.array([0, 0, 0])

    #                 elif link.ets[k].axis == 'ty':
    #                     J[:3, j] = o
    #                     J[3:, j] = np.array([0, 0, 0])

    #                 elif link.ets[k].axis == 'tz':
    #                     J[:3, j] = a
    #                     J[3:, j] = np.array([0, 0, 0])

    #                 j += 1

    #     return J

    def get_path(self, endlink=None, startlink=None):
        """
        Find a path from startlink to endlink. The endlink must come after
        the startlink (ie endlink must be further away from the base link
        of the robot than startlink) in the kinematic chain and both links
        must be a part of the same branch within the robot structure. This
        method is a work in progress while an approach which generalises
        to all applications is designed.

        :param endlink: name or reference to end-effector, defaults to None
        :type endlink: str or ELink, optional
        :param startlink: name or reference to a base link, defaults to None
        :type startlink: str or ELink, optional
        :raises ValueError: link not known or ambiguous
        :return: the path from startlink to endlink
        :rtype: list of Link
        """
        path = []
        n = 0

        endlink, startlink = self._get_limit_links(endlink, startlink)

        link = endlink

        path.append(link)
        if link.isjoint:
            n += 1

        while link != startlink:
            link = link.parent
            if link is None:
                raise ValueError(
                    f'cannot find path from {startlink.name} to'
                    f' {endlink.name}')
            path.append(link)
            if link.isjoint:
                n += 1

        path.reverse()

        return path, n

    def _get_limit_links(self, endlink=None, startlink=None):
        """
        Get and validate an end-effector, and a base link

        :param endlink: name or reference to end-effector, defaults to None
        :type endlink: str or ELink, optional
        :param startlink: name or reference to a base link, defaults to None
        :type startlink: str or ELink, optional
        :raises ValueError: link not known or ambiguous
        :raises ValueError: [description]
        :raises TypeError: unknown type provided
        :return: end-effector link, base link
        :rtype: ELink, Elink

        Helper method to find or validate an end-effector and base link.
        """
        if endlink is None:

            # if we have a gripper, use it
            if len(self.grippers) == 1:
                endlink = self.grippers[0].links[0]
            elif len(self.grippers) > 1:
                # if more than one gripper, user must choose
                raise ValueError('Must specify which gripper')

            # no grippers, use ee link if just one
            elif len(self.ee_links) == 1:
                endlink = self.ee_links[0]
            else:
                # if more than one EE, user must choose
                raise ValueError('Must specify which end-effector')
        else:
            # end effector is specified
            endlink = self._getlink(endlink)

        if startlink is None:
            startlink = self.base_link
        else:
            # start effector is specified
            startlink = self._getlink(startlink)

        return endlink, startlink

    def _getlink(self, link, default=None):
        """
        Validate reference to ELink

        :param link: link
        :type link: ELink or str
        :raises ValueError: link does not belong to this ERobot
        :raises TypeError: bad argument
        :return: link reference
        :rtype: ELink

        ``robot._getlink(link)`` is a validated reference to an ELink within
        the ERobot ``robot``.  If ``link`` is:

        -  an ``ELink`` reference it is validated as belonging to
          ``robot``.
        - a string, then it looked up in the robot's link name dictionary, and
          an ELink reference returned.
        """
        if link is None:
            link = default

        if isinstance(link, str):
            if link in self.link_dict:
                return self.link_dict[link]

            raise ValueError(f'no link named {link}')

        elif isinstance(link, ELink):
            if link in self.links:
                return link
            else:
                for gripper in self.grippers:
                    if link in gripper.links:
                        return link

                raise ValueError('link not in robot links')
        else:
            raise TypeError('unknown argument')

    def jacob0(self, q, endlink=None, startlink=None, offset=None, T=None):
        """
        Manipulator geometric Jacobian in the base frame

        :param q: Joint coordinate vector
        :type q: ndarray(n)
        :param endlink: the particular link whose velocity the Jacobian describes, defaults
            to the end-effector if only one is present
        :type endlink: str or ELink
        :param startlink: the link considered as the base frame, defaults to the robots's base frame
        :type startlink: str or ELink
        :param offset: a static offset transformation matrix to apply to the
            end of endlink, defaults to None
        :type offset: SE3, optional
        :param T: The transformation matrix of the reference point which the
            Jacobian represents with respect to the base frame. Use this to
            avoid caluclating forward kinematics to save time, defaults
            to None
        :type T: SE3, optional
        :return J: Manipulator Jacobian in the base frame
        :rtype: ndarray(6,n)

        - ``robot.jacobo(q)`` is the manipulator Jacobian matrix which maps
          joint  velocity to end-effector spatial velocity expressed in the
          end-effector frame.

        End-effector spatial velocity :math:`\nu = (v_x, v_y, v_z, \omega_x, \omega_y, \omega_z)^T`
        is related to joint velocity by :math:`{}^{E}\!\nu = \mathbf{J}_m(q) \dot{q}`.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.ETS.Puma560()
            >>> puma.jacobe([0, 0, 0, 0, 0, 0])

        .. note:: This is the geometric Jacobian as described in texts by
            Corke, Spong etal., Siciliano etal.  The end-effector velocity is
            described in terms of translational and angular velocity, not a 
            velocity twist as per the text by Lynch & Park.

        .. warning:: ``startlink`` and ``endlink`` must be on the same branch,
            with ``startlink`` closest to the base.
        """  # noqa

        if offset is None:
            offset = SE3()

        path, n = self.get_path(endlink, startlink)

        q = getvector(q, self.n)

        if T is None:
            T = self.base.inv() * \
                self.fkine(q, endlink=endlink, startlink=startlink) * offset
        T = T.A
        U = np.eye(4)
        j = 0
        J = np.zeros((6, n))

        for link in path:

            if link.isjoint:
                U = U @ link.A(q[link.jindex], fast=True)

                if link == endlink:
                    U = U @ offset.A

                Tu = np.linalg.inv(U) @ T
                n = U[:3, 0]
                o = U[:3, 1]
                a = U[:3, 2]
                x = Tu[0, 3]
                y = Tu[1, 3]
                z = Tu[2, 3]

                if link.v.axis == 'Rz':
                    J[:3, j] = (o * x) - (n * y)
                    J[3:, j] = a

                elif link.v.axis == 'Ry':
                    J[:3, j] = (n * z) - (a * x)
                    J[3:, j] = o

                elif link.v.axis == 'Rx':
                    J[:3, j] = (a * y) - (o * z)
                    J[3:, j] = n

                elif link.v.axis == 'tx':
                    J[:3, j] = n
                    J[3:, j] = np.array([0, 0, 0])

                elif link.v.axis == 'ty':
                    J[:3, j] = o
                    J[3:, j] = np.array([0, 0, 0])

                elif link.v.axis == 'tz':
                    J[:3, j] = a
                    J[3:, j] = np.array([0, 0, 0])

                j += 1
            else:
                U = U @ link.A(fast=True)

        return J

    def jacobe(self, q, endlink=None, startlink=None, offset=None, T=None):
        """
        Manipulator geometric Jacobian in the end-effector frame

        :param q: Joint coordinate vector
        :type q: ndarray(n)
        :param endlink: the particular link whose velocity the Jacobian describes, defaults
            to the end-effector if only one is present
        :type endlink: str or ELink
        :param startlink: the link considered as the base frame, defaults to the robots's base frame
        :type startlink: str or ELink
        :param offset: a static offset transformation matrix to apply to the
            end of endlink, defaults to None
        :type offset: SE3, optional
        :param T: The transformation matrix of the reference point which the
            Jacobian represents with respect to the base frame. Use this to
            avoid caluclating forward kinematics to save time, defaults
            to None
        :type T: SE3, optional
        :return J: Manipulator Jacobian in the end-effector frame
        :rtype: ndarray(6,n)

        - ``robot.jacobe(q)`` is the manipulator Jacobian matrix which maps
          joint  velocity to end-effector spatial velocity expressed in the
          end-effector frame.

        End-effector spatial velocity :math:`\nu = (v_x, v_y, v_z, \omega_x, \omega_y, \omega_z)^T`
        is related to joint velocity by :math:`{}^{E}\!\nu = \mathbf{J}_m(q) \dot{q}`.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.ETS.Puma560()
            >>> puma.jacobe([0, 0, 0, 0, 0, 0])

        .. note:: This is the **geometric Jacobian** as described in texts by
            Corke, Spong etal., Siciliano etal.  The end-effector velocity is
            described in terms of translational and angular velocity, not a 
            velocity twist as per the text by Lynch & Park.

        .. warning:: ``startlink`` and ``endlink`` must be on the same branch,
            with ``startlink`` closest to the base.
        """  # noqa

        q = getvector(q, self.n)

        if offset is None:
            offset = SE3()

        endlink, startlink = self._get_limit_links(endlink, startlink)

        path, n = self.get_path(endlink, startlink)

        if T is None:
            T = self.base.inv() * \
                self.fkine(q, endlink=endlink, startlink=startlink) * offset

        J0 = self.jacob0(q, endlink, startlink, offset, T)
        Je = self.jacobev(q, endlink, startlink, offset, T) @ J0
        return Je

    def partial_fkine0(self, q, n, J0=None, endlink=None, startlink=None):
        endlink, startlink = self._get_limit_links(endlink, startlink)

        def cross(a, b):
            x = a[1] * b[2] - a[2] * b[1]
            y = a[2] * b[0] - a[0] * b[2]
            z = a[0] * b[1] - a[1] * b[0]
            return np.array([x, y, z])

        _, nl = self.get_path(endlink, startlink)

        J = self.jacob0(q, endlink=endlink, startlink=startlink)
        H = self.hessian0(q, J, endlink, startlink)

        d = [J, H]
        size = [6, nl, nl]
        count = np.array([0, 0])
        c = 2

        def add_indices(indices, c):
            total = len(indices * 2)
            new_indices = []

            for i in range(total):
                j = i // 2
                new_indices.append([])
                new_indices[i].append(indices[j][0].copy())
                new_indices[i].append(indices[j][1].copy())

                # if even number
                if i % 2 == 0:
                    new_indices[i][0].append(c)
                # if odd number
                else:
                    new_indices[i][1].append(c)

            return new_indices

        def add_pdi(pdi):
            total = len(pdi * 2)
            new_pdi = []

            for i in range(total):
                j = i // 2
                new_pdi.append([])
                new_pdi[i].append(pdi[j][0])
                new_pdi[i].append(pdi[j][1])

                # if even number
                if i % 2 == 0:
                    new_pdi[i][0] += 1
                # if odd number
                else:
                    new_pdi[i][1] += 1

            return new_pdi

        # these are the indices used for the hessian
        indices = [[[1], [0]]]

        # the are the pd indices used in the corss prods
        pdi = [[0, 0]]

        while len(d) != n:
            size.append(nl)
            count = np.r_[count, 0]
            indices = add_indices(indices, c)
            pdi = add_pdi(pdi)
            c += 1

            pd = np.zeros(size)

            for i in range(nl ** c):

                rot = np.zeros(3)
                trn = np.zeros(3)

                for j in range(len(indices)):
                    pdr0 = d[pdi[j][0]]
                    pdr1 = d[pdi[j][1]]

                    idx0 = count[indices[j][0]]
                    idx1 = count[indices[j][1]]

                    rot += cross(
                        pdr0[(slice(3, 6), *idx0)],
                        pdr1[(slice(3, 6), *idx1)])

                    trn += cross(
                        pdr0[(slice(3, 6), *idx0)],
                        pdr1[(slice(0, 3), *idx1)])

                pd[(slice(0, 3), *count)] = trn
                pd[(slice(3, 6), *count)] = rot

                count[0] += 1
                for j in range(len(count)):
                    if count[j] == nl:
                        count[j] = 0
                        if (j != len(count) - 1):
                            count[j + 1] += 1

            d.append(pd)

        return d[-1]

    # def third(self, q=None, J0=None, endlink=None, startlink=None):
    #     endlink, startlink = self._get_limit_links(endlink, startlink)
    #     path, n = self.get_path(endlink, startlink)

    #     def cross(a, b):
    #         x = a[1] * b[2] - a[2] * b[1]
    #         y = a[2] * b[0] - a[0] * b[2]
    #         z = a[0] * b[1] - a[1] * b[0]
    #         return np.array([x, y, z])

    #     if J0 is None:
    #         q = getvector(q, n)
    #         J0 = self.jacob0(q, endlink=endlink)
    #     else:
    #         verifymatrix(J0, (6, n))

    #     H0 = self.hessian0(q, J0, endlink, startlink)

    #     L = np.zeros((6, n, n, n))

    #     for l in range(n):
    #         for k in range(n):
    #             for j in range(n):

    #                 L[:3, j, k, l] = cross(H0[3:, k, l], J0[:3, j]) + \
    #                     cross(J0[3:, k], H0[:3, j, l])

    #                 L[3:, j, k, l] = cross(H0[3:, k, l], J0[3:, j]) + \
    #                     cross(J0[3:, k], H0[3:, j, l])

    #     return L

    def hessian0(self, q=None, J0=None, endlink=None, startlink=None):
        """
        The manipulator Hessian tensor maps joint acceleration to end-effector
        spatial acceleration, expressed in the world-coordinate frame. This
        function calulcates this based on the ETS of the robot. One of J0 or q
        is required. Supply J0 if already calculated to save computation time

        :param q: The joint angles/configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param J0: The manipulator Jacobian in the 0 frame
        :type J0: float ndarray(6,n)
        :param endlink: the final link which the Hessian represents
        :type endlink: str or ELink
        :param startlink: the first link which the Hessian represents
        :type startlink: str or ELink

        :return: The manipulator Hessian in 0 frame
        :rtype: float ndarray(6,n,n)

        H[i,j,k] is d2 u_i / dq_j dq_k

        where u = {t_x, t_y, t_z, r_x, r_y, r_z}

        J[i,j] is d u_i / dq_j

        where u = {t_x, t_y, t_z, ζ_x, ζ_y, ζ_z}

        v = J qd

        a = Jd qd + J qdd

        Jd = H qd

        :references:
            - Kinematic Derivatives using the Elementary Transform
              Sequence, J. Haviland and P. Corke
        """

        endlink, startlink = self._get_limit_links(endlink, startlink)
        path, n = self.get_path(endlink, startlink)

        def cross(a, b):
            x = a[1] * b[2] - a[2] * b[1]
            y = a[2] * b[0] - a[0] * b[2]
            z = a[0] * b[1] - a[1] * b[0]
            return np.array([x, y, z])

        if J0 is None:
            q = getvector(q, n)
            J0 = self.jacob0(q, endlink=endlink)
        else:
            verifymatrix(J0, (6, n))

        H = np.zeros((6, n, n))

        for j in range(n):
            for i in range(j, n):

                H[:3, i, j] = cross(J0[3:, j], J0[:3, i])
                H[3:, i, j] = cross(J0[3:, j], J0[3:, i])

                if i != j:
                    H[:3, j, i] = H[:3, i, j]

        return H

    def jacobm(self, q=None, J=None, H=None, endlink=None, startlink=None):
        """
        Calculates the manipulability Jacobian. This measure relates the rate
        of change of the manipulability to the joint velocities of the robot.
        One of J or q is required. Supply J and H if already calculated to
        save computation time

        :param q: The joint angles/configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param J: The manipulator Jacobian in any frame
        :type J: float ndarray(6,n)
        :param H: The manipulator Hessian in any frame
        :type H: float ndarray(6,n,n)
        :param endlink: the final link which the Hessian represents
        :type endlink: str or ELink
        :param startlink: the first link which the Hessian represents
        :type startlink: str or ELink

        :return: The manipulability Jacobian
        :rtype: float ndarray(n)

        :references:
            - Kinematic Derivatives using the Elementary Transform
              Sequence, J. Haviland and P. Corke
        """

        endlink, startlink = self._get_limit_links(endlink, startlink)
        path, n = self.get_path(endlink, startlink)

        if J is None:
            if q is None:
                q = np.copy(self.q)
            else:
                q = getvector(q, self.n)

            J = self.jacob0(q, startlink=startlink, endlink=endlink)
        else:
            verifymatrix(J, (6, n))

        if H is None:
            H = self.hessian0(J0=J, startlink=startlink, endlink=endlink)
        else:
            verifymatrix(H, (6, n, n))

        manipulability = self.manipulability(
            q, J=J, startlink=startlink, endlink=endlink)
        b = np.linalg.inv(J @ np.transpose(J))
        Jm = np.zeros((n, 1))

        for i in range(n):
            c = J @ np.transpose(H[:, :, i])
            Jm[i, 0] = manipulability * \
                np.transpose(c.flatten('F')) @ b.flatten('F')

        return Jm

    def __str__(self):
        """
        Pretty prints the ETS Model of the robot. Will output angles in
        degrees

        :return: Pretty print of the robot model
        :rtype: str

        Constant links are shown in blue.
        End-effector links are prefixed with an @
        """
        table = ANSITable(
            Column("id", headalign="^"),
            Column("link", headalign="^"),
            Column("parent", headalign="^"),
            Column("joint", headalign="^"),
            Column("ETS", headalign="^", colalign="<"),
            border="thin")
        for k, link in enumerate(self):
            color = "" if link.isjoint else "<<blue>>"
            ee = "@" if link in self.ee_links else ""
            ets = link.ets()
            table.row(
                k,
                color + ee + link.name,
                link.parent.name if link.parent is not None else "-",
                link._joint_name if link.parent is not None else "",
                ets.__str__(f"q{link._jindex}"))

        s = str(table)
        s += self.configurations_str()

        return s

    def hierarchy(self):
        """
        Pretty print the robot link hierachy

        :return: Pretty print of the robot model
        :rtype: str

        Example:

        .. runblock:: pycon

            import roboticstoolbox as rtb
            robot = rtb.models.URDF.Panda()
            robot.hierarchy()

        """

        # link = self.base_link

        def recurse(link, indent=0):
            print(' ' * indent * 2, link.name)
            for child in link.child:
                recurse(child, indent+1)

        recurse(self.base_link)

    def jacobev(
            self, q, endlink=None, startlink=None,
            offset=None, T=None):
        """
        Jv = jacobev(q) is the spatial velocity Jacobian, at joint
        configuration q, which relates the velocity in the base frame to the
        velocity in the end-effector frame.

        :param q: Joint coordinate vector
        :type q: ndarray(n)
        :param endlink: the final link which the Jacobian represents
        :type endlink: str or ELink
        :param startlink: the first link which the Jacobian represents
        :type startlink: str or ELink
        :param offset: a static offset transformation matrix to apply to the
            end of endlink, defaults to None
        :type offset: SE3, optional
        :param T: The transformation matrix of the reference point which the
            Jacobian represents with respect to the base frame. Use this to
            avoid caluclating forward kinematics to save time, defaults
            to None
        :type T: SE3, optional

        :returns J: The velocity Jacobian in ee frame
        :rtype J: float ndarray(6,6)

        """

        endlink, startlink = self._get_limit_links(endlink, startlink)

        if T is None:
            T = self.base.inv() * \
                self.fkine(q, endlink=endlink, startlink=startlink)
            if offset is not None:
                T *= offset
        R = (T.R).T

        Jv = np.zeros((6, 6))
        Jv[:3, :3] = R
        Jv[3:, 3:] = R

        return Jv

    def jacob0v(
            self, q, endlink=None, startlink=None,
            offset=None, T=None):
        """
        Jv = jacob0v(q) is the spatial velocity Jacobian, at joint
        configuration q, which relates the velocity in the end-effector frame
        to velocity in the base frame

        :param q: Joint coordinate vector
        :type q: ndarray(n)
        :param endlink: the final link which the Jacobian represents
        :type endlink: str or ELink
        :param startlink: the first link which the Jacobian represents
        :type startlink: str or ELink
        :param offset: a static offset transformation matrix to apply to the
            end of endlink, defaults to None
        :type offset: SE3, optional
        :param T: The transformation matrix of the reference point which the
            Jacobian represents with respect to the base frame. Use this to
            avoid caluclating forward kinematics to save time, defaults
            to None
        :type T: SE3, optional

        :returns J: The velocity Jacobian in 0 frame
        :rtype J: float ndarray(6,6)

        """

        endlink, startlink = self._get_limit_links(endlink, startlink)

        if T is None:
            T = self.base.inv() * \
                self.fkine(q, endlink=endlink, startlink=startlink)
            if offset is not None:
                T *= offset
        R = (T.R)

        Jv = np.zeros((6, 6))
        Jv[:3, :3] = R
        Jv[3:, 3:] = R

        return Jv

    def link_collision_damper(
            self, shape, q=None, di=0.3, ds=0.05, xi=1.0,
            endlink=None, startlink=None):
        '''
        Formulates an inequality contraint which, when optimised for will
        make it impossible for the robot to run into a collision. Requires
        See examples/neo.py for use case

        :param ds: The minimum distance in which a joint is allowed to
            approach the collision object shape
        :type ds: float
        :param di: The influence distance in which the velocity
            damper becomes active
        :type di: float
        :param xi: The gain for the velocity damper
        :type xi: float
        :param from_link: The first link to consider, defaults to the base
            link
        :type from_link: ELink
        :param to_link: The last link to consider, will consider all links
            between from_link and to_link in the robot, defaults to the
            end-effector link
        :type to_link: ELink

        :returns: Ain, Bin as the inequality contraints for an omptimisor
        :rtype: ndarray(6), ndarray(6)
        '''

        if startlink is None:
            startlink = self.base_link

        if endlink is None:
            endlink = self.ee_link

        links, n = self.get_path(startlink=startlink, endlink=endlink)

        if q is None:
            q = np.copy(self.q)
        else:
            q = getvector(q, n)

        j = 0
        Ain = None
        bin = None

        def indiv_calculation(link, link_col, q):
            d, wTlp, wTcp = link_col.closest_point(shape, di)

            if d is not None:
                lpTcp = wTlp.inv() * wTcp
                norm = lpTcp.t / d
                norm_h = np.expand_dims(np.r_[norm, 0, 0, 0], axis=0)

                Je = self.jacobe(
                    q, startlink=self.base_link, endlink=link,
                    offset=link_col.base)
                n_dim = Je.shape[1]
                dp = norm_h @ shape.v
                l_Ain = np.zeros((1, n))
                l_Ain[0, :n_dim] = norm_h @ Je
                l_bin = (xi * (d - ds) / (di - ds)) + dp
            else:
                l_Ain = None
                l_bin = None

            return l_Ain, l_bin, d, wTcp

        for link in links:
            if link.isjoint:
                j += 1

            for link_col in link.collision:
                l_Ain, l_bin, d, wTcp = indiv_calculation(
                                                link, link_col, q)

                if l_Ain is not None and l_bin is not None:
                    if Ain is None:
                        Ain = l_Ain
                    else:
                        Ain = np.r_[Ain, l_Ain]

                    if bin is None:
                        bin = np.array(l_bin)
                    else:
                        bin = np.r_[bin, l_bin]

        return Ain, bin

    # inverse dynamics (recursive Newton-Euler) using spatial vector notation
    def rne(robot, q, qd, qdd, gravity=None):

        n = robot.n

        # allocate intermediate variables
        Xup = SE3.Alloc(n)
        Xtree = SE3.Alloc(n)

        v = SpatialVelocity.Alloc(n)
        a = SpatialAcceleration.Alloc(n)
        f = SpatialForce.Alloc(n)
        I = SpatialInertia.Alloc(n)  # noqa
        s = [None for i in range(n)]   # joint motion subspace
        Q = np.zeros((n,))   # joint torque/force

        # initialize intermediate variables
        for j, link in enumerate(robot):
            I[j] = SpatialInertia(m=link.m, r=link.r)
            Xtree[j] = link.Ts
            s[j] = link.v.s

        if gravity is None:
            a_grav = SpatialAcceleration(robot.gravity)
        else:
            a_grav = SpatialAcceleration(gravity)

        # forward recursion
        for j in range(0, n):
            vJ = SpatialVelocity(s[j] * qd[j])

            # transform from parent(j) to j
            Xup[j] = robot[j].A(q[j]).inv()

            if robot[j].parent is None:
                v[j] = vJ
                a[j] = Xup[j] * a_grav + SpatialAcceleration(s[j] * qdd[j])
            else:
                jp = robot[j].parent.jindex
                v[j] = Xup[j] * v[jp] + vJ
                a[j] = Xup[j] * a[jp] \
                    + SpatialAcceleration(s[j] * qdd[j]) \
                    + v[j] @ vJ

            f[j] = I[j] * a[j] + v[j] @ (I[j] * v[j])

        # backward recursion
        for j in reversed(range(0, n)):
            Q[j] = f[j].dot(s[j])

            if robot[j].parent is not None:
                jp = robot[j].parent.jindex
                f[jp] = f[jp] + Xup[j] * f[j]

        return Q


if __name__ == "__main__":  # pragma nocover

    import roboticstoolbox as rtb
    np.set_printoptions(precision=4, suppress=True)

    p = rtb.models.ETS.Puma560()
    p.fkine(p.qz)
    p.jacob0(p.qz)
    p.jacobe(p.qz)

    # robot = rtb.models.ETS.Panda()
    # print(robot)
    # print(robot.base, robot.tool)
    # print(robot.ee_links)
    # ets = robot.ets()
    # print(ets)
    # print('n', ets.n)
    # ets2 = ets.compile()
    # print(ets2)

    # q = np.random.rand(7)
    # # print(ets.eval(q))
    # # print(ets2.eval(q))

    # J1 = robot.jacob0(q)
    # J2 = ets2.jacob0(q)
    # print(J1-J2)

    # print(robot[2].v, robot[2].v.jindex)
    # print(robot[2].Ts)



"""
Python Vehicle
@Author: Kristian Gibson
@Author: Peter Corke
"""
from abc import ABC, abstractmethod
import warnings
from math import pi, sin, cos, tan, atan2
import numpy as np
from scipy import integrate, linalg, interpolate

import matplotlib.pyplot as plt
from matplotlib import patches
import matplotlib.transforms as mtransforms

from spatialmath import SE2, base
import roboticstoolbox as rtb


class VehicleDriver(ABC):

    @abstractmethod
    def demand():
        pass

    @abstractmethod
    def init():
        pass

    @abstractmethod
    def vehicle():
        pass    

# ========================================================================= #

class RandomPath(VehicleDriver):
    """
        RandomPath Vehicle driver class

    Create a "driver" object capable of steering a Vehicle subclass object through random 
    waypoints within a rectangular region and at constant speed.

    The driver object is connected to a Vehicle object by the latter's
    add_driver() method.  The driver's demand() method is invoked on every
    call to the Vehicle's step() method.

    Methods::
    init       reset the random number generator
    demand     speed and steer angle to next waypoint
    display    display the state and parameters in human readable form
    char       convert to string
    plot      
    Properties::
    goal          current goal/waypoint coordinate
    veh           the Vehicle object being controlled
    dim           dimensions of the work space (2x1) [m]
    speed         speed of travel [m/s]
    dthresh       proximity to waypoint at which next is chosen [m]

    Example::

    veh = Bicycle(V);
    veh.add_driver( RandomPath(20, 2) );

    Notes::
    - It is possible in some cases for the vehicle to move outside the desired
    region, for instance if moving to a waypoint near the edge, the limited
    turning circle may cause the vehicle to temporarily move outside.
    - The vehicle chooses a new waypoint when it is closer than property
    closeenough to the current waypoint.
    - Uses its own random number stream so as to not influence the performance
    of other randomized algorithms such as path planning.

    Reference::

    Robotics, Vision & Control, Chap 6,
    Peter Corke,
    Springer 2011

    See also Vehicle, Bicycle, Unicycle.

    """

    def __init__(self, dim, speed=1, dthresh=0.05, seed=None, headinggain=1, goalmarkerstyle=None):
        """
        Driving agent for random path

        :param dim: dimension of workspace, see spatialmath.plotvol2
        :type dim: scalar, array_like(2), array_like(4)
        :param speed: forward speed, defaults to 1
        :type speed: float, optional
        :param dthresh: distance threshold, defaults to 0.05
        :type dthresh: float, optional

        :raises ValueError: [description]

        Returns a *driver* object that drives the attached vehicle to a 
        sequence of random waypoints.

        The driver is connected to the vehicle by::

            Vehicle(control=driver)
            
            veh = Vehicle()
            veh.control = driver
        
        The waypoints are positioned inside a rectangular region defined by
        the vehicle that is controlled.
          - D scalar; X: -D to +D, Y: -D to +D
              - D (1x2); X: -D(1) to +D(1), Y: -D(2) to +D(2)
        %     - D (1x4); X: D(1) to D(2), Y: D(3) to D(4)
        
        % Options::
        % 'speed',S      Speed along path (default 1m/s).
        % 'dthresh',D    Distance from goal at which next goal is chosen.
        %
        % See also Vehicle.
        :seealso: :func:`spatialmath.plotvol2`
        """
        
        # TODO options to specify region, maybe accept a Map object?
        

        self._dim = base.expand_dims(dim)
        
        self._speed = speed
        self._dthresh = dthresh * np.diff(self._dim[0:2])
        self._goal_marker = None
        if goalmarkerstyle is None:
            self._goal_marker_style = {
                'marker': 'D',
                'markersize': 6, 
                'color': 'r',
                }
        else:
            self._goal_marker_style = goalmarkerstyle
        self._headinggain = headinggain
        
        self._d_prev = np.inf
        self._rand = np.random.default_rng(seed)
        self.verbose = True
        self._goal = None
        self._dthresh = dthresh * max(
                self._dim[1] - self._dim[0], 
                self._dim[3] - self._dim[2]
                                    )

        self._veh = None

    def __str__(self):
        """%RandomPath.char Convert to string
        %
        % s = R.char() is a string showing driver parameters and state in in 
        % a compact human readable format. """

        s = 'RandomPath driver object\n'
        s += f"  X {self._dim[0]} : {self._dim[1]}; Y {self._dim[0]} : {self._dim[1]}, dthresh={self._dthresh}\n"
        s += f"  current goal={self._goal}"
        return s

    @property
    def vehicle(self):
        return self._veh

    @vehicle.setter
    def vehicle(self, v):
        """
        Connect to the vehicle under control

        :param v: [description]
        :type v: [type]
        """
        self._veh = v

    def init(self, ax=None):
        """
        Initialize random path driving agent
        
        %RandomPath.init Reset random number generator
        %
        % R.init() resets the random number generator used to create the waypoints.
        % This enables the sequence of random waypoints to be repeated.
        %
        % Notes::
        % - Called by Vehicle.run.
        %
        % See also RANDSTREAM.
        """
        self._goal = None
        # delete(driver.h_goal);   % delete the goal
        # driver.h_goal = [];
        if ax is not None:
            self._goal_marker = plt.plot(np.nan, np.nan, **self._goal_marker_style)[0]

    def demand(self):
        """
        Compute speed and heading for random waypoint
            %
            % [SPEED,STEER] = R.demand() is the speed and steer angle to
            % drive the vehicle toward the next waypoint.  When the vehicle is
            % within R.dtresh a new waypoint is chosen.
            %
            % See also Vehicle."""

        if self._goal is None:
            self._new_goal()

        # if nearly at goal point, choose the next one
        d = np.linalg.norm(self._veh._x[0:2] - self._goal)
        if d < self._dthresh:
            self._new_goal()
        # elif d > 2 * self._d_prev:
        #     self.choose_goal()
        # self._d_prev = d

        speed = self._speed

        goal_heading = atan2(
            self._goal[1] - self._veh._x[1], 
            self._goal[0] - self._veh._x[0]
                )
        delta_heading = base.angdiff(goal_heading, self._veh._x[2])

        return np.r_[speed, self._headinggain * delta_heading]

    ## private method, invoked from demand() to compute a new waypoint
    
    def _new_goal(self):
        
        # choose a uniform random goal within inner 80% of driving area
        while True:
            r = self._rand.uniform(0.1, 0.9)
            gx = self._dim[0:2] @ np.r_[r, 1-r]

            r = self._rand.uniform(0.1, 0.9)
            gy = self._dim[2:4] @ np.r_[r, 1-r]

            self._goal = np.r_[gx, gy]

            # check not too close to last goal
            if np.linalg.norm(self._goal - self._veh._x[0:2]) > 2 * self._dthresh:
                break

        if self._veh.verbose:
            print(f"set goal: {self._goal}")

        # update the goal marker
        if self._goal_marker is not None:
            self._goal_marker.set_xdata(self._goal[0])
            self._goal_marker.set_ydata(self._goal[1])

# ========================================================================= #

class PurePursuit(VehicleDriver):

    def __init__(self, speed=1, radius=1):
        pass

    def __str__(self):
        pass

    def init(self):
        pass

    def demand(self):
        pass

# ========================================================================= #

if __name__ == "__main__":

    import unittest





#!/usr/bin/env python
"""
@author: Jesse Haviland
"""

from spatialmath import SE3
from spatialmath.base.argcheck import getvector
from spatialmath.base import r2q
import numpy as np
from functools import wraps
from io import StringIO
import os

_mpl = False

try:
    from matplotlib import colors as mpc
    _mpl = True
except ImportError:    # pragma nocover
    pass

p = None
_pyb = None


def _import_pyb():
    import importlib
    global _pyb
    global p

    try:
        from roboticstoolbox.tools.stdout_supress import pipes
    except Exception:  # pragma nocover
        from contextlib import contextmanager

        @contextmanager
        def pipes(stdout=None, stderr=None):
            pass

    try:
        out = StringIO()
        try:
            with pipes(stdout=out, stderr=None):
                p = importlib.import_module('pybullet')
        except Exception:  # pragma nocover
            p = importlib.import_module('pybullet')

        cid = p.connect(p.SHARED_MEMORY)
        if (cid < 0):
            p.connect(p.DIRECT)
        _pyb = True
    except ImportError:   # pragma nocover
        _pyb = False


class Shape(object):

    def __init__(
            self,
            base=None,
            radius=0,
            length=0,
            scale=[1, 1, 1],
            color=None,
            filename=None,
            stype=None):

        self._wT = SE3()
        self.co = None
        self.base = base
        self.scale = scale
        self.radius = radius
        self.length = length
        self.filename = filename
        self.stype = stype
        self.v = np.zeros(6)
        self.color = color

        self.pinit = False

    def to_dict(self):
        '''
        to_dict() returns the shapes information in dictionary form

        :returns: All information about the shape
        :rtype: dict
        '''

        if self.stype == 'cylinder':
            fk = self.wT * SE3.Rx(np.pi/2)
        else:
            fk = self.wT

        shape = {
            'stype': self.stype,
            'scale': self.scale.tolist(),
            'filename': self.filename,
            'radius': self.radius,
            'length': self.length,
            't': fk.t.tolist(),
            'q': r2q(fk.R).tolist(),
            'v': self.v.tolist(),
            'color': list(self.color)
        }

        return shape

    def fk_dict(self):
        '''
        fk_dict() outputs shapes pose in dictionary form

        :returns: The shape pose in translation and quternion form
        :rtype: dict
        '''

        if self.stype == 'cylinder':
            fk = self.wT * SE3.Rx(np.pi/2)
        else:
            fk = self.wT

        shape = {
            't': fk.t.tolist(),
            'q': r2q(fk.R).tolist()
        }

        return shape

    def __repr__(self):   # pragma nocover
        return f'{self.stype},\n{self.base}'

    def _update_pyb(self):
        if _pyb and self.co is not None:
            q = r2q(self.wT.R)
            rot = [q[1], q[2], q[3], q[0]]
            p.resetBasePositionAndOrientation(self.co, self.wT.t, rot)

    def _init_pob(self):   # pragma nocover
        pass

    def _check_pyb(func):   # pragma nocover
        @wraps(func)
        def wrapper_check_pyb(*args, **kwargs):
            if _pyb is None:
                _import_pyb()
            return func(*args, **kwargs)
        return wrapper_check_pyb

    @property
    def v(self):
        return self._v

    @v.setter
    def v(self, value):
        self._v = getvector(value, 6)

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):

        default_color = (0.95, 0.5, 0.25, 1.0)

        if isinstance(value, tuple):
            value = list(value)

        if isinstance(value, list):
            if len(value) == 3:
                value.append(1.0)
                value = tuple(value)
        elif isinstance(value, str):
            if _mpl:
                try:
                    value = mpc.to_rgba(value)
                except ValueError:
                    print(f'{value} is an invalid color name, using default color')
                    value = default_color
            else:  # pragma nocover
                value = default_color
                print(
                    'Color only supported when matplotlib is installed\n'
                    'Install using: pip install matplotlib')
        elif value is None:
            value = default_color

        self._color = value

    @property
    def wT(self):
        return self._wT * self.base

    @wT.setter
    def wT(self, T):
        if not isinstance(T, SE3):
            T = SE3(T)
        self._wT = T
        self._update_pyb()

    @property
    def base(self):
        return self._base

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        if value is not None:
            value = getvector(value, 3)
        else:
            value = getvector([1, 1, 1], 3)
        self._scale = value

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = value

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = float(value)

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, value):
        self._length = float(value)

    @base.setter
    def base(self, T):
        if not isinstance(T, SE3):
            T = SE3(T)
        self._base = T
        self._update_pyb()

    @_check_pyb
    def closest_point(self, shape, inf_dist=1.0):
        '''
        closest_point(shape, inf_dist) returns the minimum euclidean
        distance between self and shape, provided it is less than inf_dist.
        It will also return the points on self and shape in the world frame
        which connect the line of length distance between the shapes. If the
        distance is negative then the shapes are collided.

        :param shape: The shape to compare distance to
        :type shape: Shape
        :param inf_dist: The minimum distance within which to consider
            the shape
        :type inf_dist: float
        :returns: d, p1, p2 where d is the distance between the shapes,
            p1 and p2 are the points in the world frame on the respective
            shapes
        :rtype: float, SE3, SE3
        '''

        if not self.pinit:
            self._init_pob()
            self._update_pyb()

        if not shape.pinit:
            shape._init_pob()
            shape._update_pyb()

        if not _pyb:  # pragma nocover
            raise ImportError(
                'The package PyBullet is required for collision '
                'functionality. Install using pip install pybullet')

        ret = p.getClosestPoints(self.co, shape.co, inf_dist)

        if len(ret) == 0:
            d = None
            p1 = None
            p2 = None
        else:
            ret = ret[0]
            d = ret[8]
            p1 = SE3(ret[5])
            p2 = SE3(ret[6])

        return d, p1, p2

    def collided(self, shape):
        '''
        collided(shape) checks if self and shape have collided

        :param shape: The shape to compare distance to
        :type shape: Shape
        :returns: True if shapes have collided
        :rtype: bool
        '''

        d, _, _ = self.closest_point(shape)

        if d is not None and d <= 0:
            return True
        else:
            return False


class Mesh(Shape):
    """
    A mesh object described by an stl or collada file.

    :param filename: The path to the mesh that contains this object.
        This is the absolute path.
    :type filename: str
    :param scale: The scaling value for the mesh along the XYZ axes. If
        ``None``, assumes no scale is applied.
    :type scale: list (3) float, optional
    :param base: Local reference frame of the shape
    :type base: SE3

    """

    def __init__(self, filename=None, scale=[1, 1, 1], base=None):
        super(Mesh, self).__init__(
            filename=filename, base=base,
            scale=scale, stype='mesh')

    def _init_pob(self):
        name, file_extension = os.path.splitext(self.filename)
        if (file_extension == '.stl' or file_extension == '.STL'):

            col = p.createCollisionShape(
                shapeType=p.GEOM_MESH,
                fileName=self.filename,
                meshScale=self.scale)

            self.co = p.createMultiBody(
                baseMass=1,
                baseInertialFramePosition=[0, 0, 0],
                baseCollisionShapeIndex=col)

            self.pinit = True


class Cylinder(Shape):
    """A cylinder whose center is at the local origin.
    Parameters

    :param radius: The radius of the cylinder in meters.
    :type radius: float
    :param length: The length of the cylinder in meters.
    :type length: float
    :param base: Local reference frame of the shape
    :type base: SE3

    """

    def __init__(self, radius, length, base=None):
        super(Cylinder, self).__init__(
            base=base, radius=radius, length=length,
            stype='cylinder')

    def _init_pob(self):
        col = p.createCollisionShape(
            shapeType=p.GEOM_CYLINDER,
            radius=self.radius, height=self.length)

        self.co = p.createMultiBody(
            baseMass=1,
            baseInertialFramePosition=[0, 0, 0],
            baseCollisionShapeIndex=col)

        self.pinit = True


class Sphere(Shape):
    """
    A sphere whose center is at the local origin.

    :param radius: The radius of the sphere in meters.
    :type radius: float
    :param base: Local reference frame of the shape
    :type base: SE3

    """

    def __init__(self, radius, base=None):
        super(Sphere, self).__init__(
            base=base, radius=radius, stype='sphere')

    def _init_pob(self):
        col = p.createCollisionShape(
            shapeType=p.GEOM_SPHERE, radius=self.radius)

        self.co = p.createMultiBody(
            baseMass=1,
            baseInertialFramePosition=[0, 0, 0],
            baseCollisionShapeIndex=col)

        self.pinit = True


class Box(Shape):
    """
    A rectangular prism whose center is at the local origin.

    :param scale: The length, width, and height of the box in meters.
    :type scale: list (3) float
    :param base: Local reference frame of the shape
    :type base: SE3

    """

    def __init__(self, scale, base=None):
        super(Box, self).__init__(base=base, scale=scale, stype='box')

    def _init_pob(self):

        col = p.createCollisionShape(
            shapeType=p.GEOM_BOX,
            halfExtents=np.array(self.scale)/2)

        self.co = p.createMultiBody(
            baseMass=1,
            baseInertialFramePosition=[0, 0, 0],
            baseCollisionShapeIndex=col)

        self.pinit = True
